r'''.
           @@@@@@@@@@
       @@@@..........@@@@
    @@@         .        @@@
  @@.           .         . @@
 @  .     _     .         .   @
@........| |...................@    *********************************************
@      . | |   _____  .        @
@      . | |  |  __ \ .        @    La Data Web
@      . | |__| |  | |.   ***  @
@........|____| |  | |...*   *.@    Copyright © 2025 Ignacio Barrau
@   .       . | |__| |. *     *@
@   .       . |_____/ . *     *@    *********************************************
@   .       .         . *     *@
@   .       .         . *******@
@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
'''

import json
import requests
from simplepbi import utils
import pandas as pd
from simplepbi.fabric.core import LongRunningOperations
import re
import base64
class SemanticModels():
    """Simple library to use the api and obtain semantic models item from it.
    """

    def __init__(self, token):
        """Create a simplePBI object to request fabric semantic models item API
        Args:
            token: String
                Bearer Token to use the Rest API
        """
        self.token = token

    # Get semantic model in Workspace
    def get_semantic_model(self, workspace_id, semantic_model_id):
        """Returns properties of the specified semantic model.
        ### Parameters
        ----
        workspace_id: str uuid
            The workspace id. You can take it from PBI Service URL
        semantic_model_id: str uuid
            The semantic model id. You can take it from PBI Service URL
        ### Returns
        ----
        Dict:
            A dictionary containing a semantic model in the workspace.
        """
        try:
            url = "https://api.fabric.microsoft.com/v1/workspaces/{}/semanticModels/{}".format(workspace_id, semantic_model_id)
            res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)

    # List semantic models in Workspace
    def list_semantic_models(self, workspace_id):
        """Returns a list of semantic models in the specified workspace.
        ### Parameters
        ----
        workspace_id: str uuid
            The workspace id. You can take it from PBI Service URL
        ### Returns
        ----
        Dict:
            A dictionary containing all the semantic models in the workspace.
        """
        try:
            url = "https://api.fabric.microsoft.com/v1/workspaces/{}/semanticModels".format(workspace_id)
            res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)

    # Delete semantic model in Workspace
    def delete_semantic_model(self, workspace_id, semantic_model_id):
        """Deletes the specified semantic model from the specified workspace.
        ### Parameters
        ----
        workspace_id: str uuid
            The workspace id. You can take it from PBI Service URL
        semantic_model_id: str uuid
            The semantic model id. You can take it from PBI Service URL
        ### Returns
        ----
        Response object from requests library. 200 OK
        """
        try:
            url = "https://api.fabric.microsoft.com/v1/workspaces/{}/semanticModels/{}".format(workspace_id, semantic_model_id)
            headers = {'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)}
            res = requests.delete(url, headers=headers)
            res.raise_for_status()
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
    # Create semantic model in Workspace
    
    def get_semantic_model_definition(self, workspace_id, semantic_model_id, format=None):
        """Returns the definition of the specified semantic model.
        ### Parameters
        ----
        workspace_id: str uuid
            The workspace id. You can take it from PBI Service URL
        semantic_model_id: str uuid
            The semantic model id. You can take it from PBI Service URL
        format: str
            The format of the semantic model definition. Can be 'TMDL' or 'TMSL'.
        ### Returns
        ----
        Dict:
            A dictionary containing the definition of the semantic model.
        """
        try:
            op = LongRunningOperations(self.token)
            url = "https://api.fabric.microsoft.com/v1/workspaces/{}/semanticModels/{}/getDefinition".format(workspace_id, semantic_model_id)
            if format != None:
                url += "?format={}".format(format)
            res = requests.post(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            opid = res.headers["x-ms-operation-id"]
            status="Running"
            while status == "Running":
                try:                    
                    ope = op.get_operation_state(opid)
                    status = json.loads(ope)["status"]                    
                except Exception as e:
                    print("Error while checking operation status: ", e)
                    break
            model = op.get_operation_result(opid)
            print("Operation completed with status: ", status)
            return json.loads(model)
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)

    def list_roles_from_semantic_model(self, workspace_id, semantic_model_id, model=None):
        """Returns the roles of the specified semantic model.
        ### Parameters
        ----
        workspace_id: str uuid
            The workspace id. You can take it from PBI Service URL
        semantic_model_id: str uuid
            The semantic model id. You can take it from PBI Service URL
        model: object
            TMDL model definition, if you already have it, you don't need to specify workspace_id and semantic_model_id. Send empty instead.
        ### Returns
        ----
        DataFrame:
            A list containing the roles of the semantic model.
        """
        try:
            if model is None:
                model = self.get_semantic_model_definition(workspace_id, semantic_model_id,"TMDL")
            for i in model["definition"]["parts"]:
                if "model.tmdl" in i["path"]:
                    metad = i["payload"]
            decoded_contents = base64.b64decode(metad.encode("utf-8"))
            decoded_str = decoded_contents.decode("utf-8")
            roles = [role.strip().strip("'\"") for role in re.findall(r'^\s*ref role (.+)$', decoded_str, re.MULTILINE)]
            if roles == []:
                print("No roles found in the semantic model.")
            return roles
        except Exception as e:
            print("Error while getting roles: ", e)
    def list_tables_from_semantic_model(self, workspace_id, semantic_model_id, model=None):
        """Returns the tables of the specified semantic model.
        ### Parameters
        ----
        workspace_id: str uuid
            The workspace id. You can take it from PBI Service URL
        semantic_model_id: str uuid
            The semantic model id. You can take it from PBI Service URL
        model: object
            TMDL model definition, if you already have it, you don't need to specify workspace_id and semantic_model_id. Send empty instead.
        ### Returns
        ----
        DataFrame:
            A list containing the tables of the semantic model.
        """
        try:
            if model is None:
                model = self.get_semantic_model_definition(workspace_id, semantic_model_id,"TMDL")
            for i in model["definition"]["parts"]:
                if "model.tmdl" in i["path"]:
                    metad = i["payload"]
            decoded_contents = base64.b64decode(metad.encode("utf-8"))
            decoded_str = decoded_contents.decode("utf-8")
            tablas = [tabla.strip().strip("'\"") for tabla in re.findall(r'^\s*ref table (.+)$', decoded_str, re.MULTILINE)]
            if tablas == []:
                print("No tables found in the semantic model.")
            return tablas
        except Exception as e:
            print("Error while getting tables: ", e)
    
    def get_tables_schema_from_semantic_model(self, workspace_id, semantic_model_id, model=None):
        """Returns the tables schema of the specified semantic model.
        ### Parameters
        ----
        workspace_id: str uuid
            The workspace id. You can take it from PBI Service URL
        semantic_model_id: str uuid
            The semantic model id. You can take it from PBI Service URL
        model: object
            TMDL model definition, if you already have it, you don't need to specify workspace_id and semantic_model_id. Send empty instead.
        ### Returns
        ----
        DataFrame:
            A pandas DataFrame containing the tables of the semantic model.
        """
        try:
            if model is None:
                model = self.get_semantic_model_definition(workspace_id, semantic_model_id, "TMDL")
            tables = [base64.b64decode(i["payload"].encode("utf-8")).decode("utf-8") for i in model["definition"]["parts"] if "tables" in i["path"]]
            # Append all tables into a single string
            decoded_str = "\n".join(tables)
            return utils.parse_tmdl_structure(decoded_str)
        except Exception as e:
            print("Error while getting tables: ", e)

    def get_tables_partitions_from_semantic_model(self, workspace_id, semantic_model_id, model=None):
        """Returns the tables partitions of the specified semantic model.
        ### Parameters
        ----
        workspace_id: str uuid
            The workspace id. You can take it from PBI Service URL
        semantic_model_id: str uuid
            The semantic model id. You can take it from PBI Service URL
        model: object
            TMDL model definition, if you already have it, you don't need to specify workspace_id and semantic_model_id. Send empty instead.
        ### Returns
        ----
        DataFrame:
            A pandas DataFrame containing the tables of the semantic model.
        """
        try:
            if model is None:
                model = self.get_semantic_model_definition(workspace_id, semantic_model_id, "TMDL")
            tables = [base64.b64decode(i["payload"].encode("utf-8")).decode("utf-8") for i in model["definition"]["parts"] if "tables" in i["path"]]
            # Append all tables into a single string
            decoded_str = "\n".join(tables)             
            # Find all tables and their partitions with M code
            table_blocks = re.findall(
                r'table\s+([^\s]+)(.*?)partition\s+\1-[\w-]+.*?=\s*m\s+.*?source\s*=\s*(let\s+.*?in\s+.*?)(?=\n\t*annotation|\n\s*$)',
                decoded_str,
                flags=re.DOTALL,
            )

            # Build list of dicts
            records = []
            for table_name, _, m_code in table_blocks:
                records.append({
                    "table_name": table_name.strip(),
                    "m_code": m_code.strip()
                })
            df = pd.DataFrame(records)
            df.rename(columns={"table_name": "table", "m_code": "query_definition"}, inplace=True)
            return df
        except Exception as e:
            print("Error while getting tables: ", e)
    
    def get_relationships_from_semantic_model(self, workspace_id, semantic_model_id, model=None):
        """Returns the relationships of the specified semantic model.
        ### Parameters
        ----
        workspace_id: str uuid
            The workspace id. You can take it from PBI Service URL
        semantic_model_id: str uuid
            The semantic model id. You can take it from PBI Service URL
        model: object
            TMDL model definition, if you already have it, you don't need to specify workspace_id and semantic_model_id. Send empty instead.
        ### Returns
        ----
        DataFrame:
            A pandas DataFrame containing the relationships of the semantic model.
        """
        try:
            if model is None:
                model = self.get_semantic_model_definition(workspace_id, semantic_model_id, "TMDL")
            for i in model["definition"]["parts"]:
                if "relationships.tmdl" in i["path"]:
                    relations = i["payload"]
            relationships = base64.b64decode(relations.encode("utf-8")).decode("utf-8")
            return utils.extract_relationships(relationships)
        except Exception as e:
            print("Error while getting relationships: ", e)

    def get_semantic_model_bim(self, workspace_id, semantic_model_id):
        """Returns the model.bim of the specified semantic model.
        ### Parameters
        ----
        workspace_id: str uuid
            The workspace id. You can take it from PBI Service URL
        semantic_model_id: str uuid
            The semantic model id. You can take it from PBI Service URL
        ### Returns
        ----
        DataFrame:
            A pandas DataFrame containing the bim of the semantic model.
        """
        try:
            model = self.get_semantic_model_definition(workspace_id, semantic_model_id, "TMSL")
            for i in model["definition"]["parts"]:
                if "model.bim" in i["path"]:
                    datamodel = i["payload"]
            decoded_contents = base64.b64decode(datamodel.encode("utf-8"))            
            return json.loads(decoded_contents)
        except Exception as e:
            print("Error while getting bim file: ", e)

    def create_html_semantic_model_documentation(self, workspace_id, semantic_model_id, output_html_path):
        """Generates an HTML documentation for the specified semantic model.
        ### Parameters
        ----
        workspace_id: str uuid
            The workspace id. You can take it from PBI Service URL
        semantic_model_id: str uuid
            The semantic model id. You can take it from PBI Service URL
        output_html_path: str
            The path where the HTML documentation will be saved.
            E.g: C:\\Users\\user\\Desktop\\semantic_model_documentation.html
        ### Returns
        ----
        None
        """
        try:
            print("Getting semantic model info...")
            bim = self.get_semantic_model_bim(workspace_id, semantic_model_id)            
            print("Generating file...")
            utils.generate_bim_documentation(json.dumps(bim), output_html_path)            
            print("Documentation generated successfully at: ", output_html_path)
        except Exception as e:
            print("Error while generating HTML documentation: ", e)


r'''.
           @@@@@@@@@@
       @@@@..........@@@@
    @@@         .        @@@
  @@.           .         . @@
 @  .     _     .         .   @
@........| |...................@    *********************************************
@      . | |   _____  .        @
@      . | |  |  __ \ .        @    La Data Web
@      . | |__| |  | |.   ***  @
@........|____| |  | |...*   *.@    Copyright © 2022 Ignacio Barrau
@   .       . | |__| |. *     *@
@   .       . |_____/ . *     *@    *********************************************
@   .       .         . *     *@
@   .       .         . *******@
@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
'''

import json
import pandas as pd
import io
import re
from typing import Dict, Any

'''
class Utils():
    """Simple library class."""
    
    def __init__(self):
        """Class with useful methods."""
    '''
    
def append_value(dict_obj, key, value):
    '''
    Parameters
    ----------
    dict_obj : dict
        dictionary object to append values.
    key : str
        Key name of the dictonary that will append a value.
    value : any
        Value to append in the key of the dictionary.

    Returns
    -------
    None. It is applied to the dict in the parameter
    '''
    # Check if key exist in dict or not
    if key in dict_obj:
        # Key exist in dict.
        # Check if type of value of key is list or not
        if not isinstance(dict_obj[key], list):
            # If type is not list then make it list
            dict_obj[key] = [dict_obj[key]]
            # Append the value in list
            dict_obj[key].append(value)
        else:
            # As key is not in dict,
            # so, add key-value pair
            dict_obj[key] = value
    
def to_pandas(response_dict, father_node):
    '''
    Parameters
    ----------
    response_dict : dict
        response from methods.
    father_node : TYPE
        Dict Key involving the list [] of items. Example for get_dashboards, the key of the response is "value"
        {
          "value": [
            {
              "id": "69ffaa6c-b36d-4d01-96f5-1ed67c64d4af",
              "displayName": "SalesMarketing",
              "embedUrl": "https://app.powerbi.com/dashboardEmbed?dashboardId=69ffaa6c-b36d-4d01-96f5-1ed67c64d4af&groupId=f089354e-8366-4e18-aea3-4cb4a3a50b48",
              "isReadOnly": false
            }
          ]
        }

    Returns
    -------
    df : Pandas DataFrame
        Returns the dict response converted in a Pandas DataFrame.
    '''
    js = json.dumps(response_dict[father_node])
    df = pd.read_json(io.StringIO(js))
    return df 

def get_artifact_from_scan_preview(scan_result, artifact):
    """Get a table of an specific artifact
    ### Parameters
    ----
    scan_result: dict
        The scan result response from get_scan_result_preview
    artifact: str
        The type of artifact. Types: 'reports', 'dashboards', 'datasets', 'dataflows', 'users'
    ### Returns
    ----
    str:
        Returns the status of the scan. Succeeded means you are ready to request scans.
    """
    df_total = pd.DataFrame()
    try:        
        for group in scan_result["workspaces"]:
            df = pd.read_json(json.dumps(group[artifact]))
            df["workspaceId"] = group["id"]
            df_total = pd.concat([df_total, df], sort=True, ignore_index=True)
        return df_total
    except Exception as e:
        print("ERROR: ", e)

def extract_relationships(relationships_text):
    """
    Parses TMDL-formatted relationship definitions into a structured DataFrame.
    ### Parameters
    ----
    relationships_text: str
        The TMDL-formatted string containing relationship definitions.
    ### Returns
    ----
    DataFrame:
        A pandas DataFrame with columns: fromTable, fromColumn, toTable, toColumn.
    """
    relationships = []
    current_from = None
    current_to = None

    lines = relationships_text.splitlines()
    for line in lines:
        line = line.strip()

        if line.startswith("fromColumn:"):
            from_full = line.split(":", 1)[1].strip()
            if '.' in from_full:
                from_table, from_column = from_full.split('.', 1)
            else:
                from_table, from_column = None, from_full

        elif line.startswith("toColumn:"):
            to_full = line.split(":", 1)[1].strip()
            if '.' in to_full:
                to_table, to_column = to_full.split('.', 1)
            else:
                to_table, to_column = None, to_full

            # Append relationship once both from and to are found
            relationships.append({
                "fromTable": from_table.strip("'\""),
                "fromColumn": from_column.strip("'\""),
                "toTable": to_table.strip("'\""),
                "toColumn": to_column.strip("'\"")
            })

    return pd.DataFrame(relationships)    

def parse_tmdl_structure(table_definition):
    """
    Parses a TMDL-format semantic model tables definition into a structured DataFrame.
    Extracts table name, field name, data type, and DAX expression.
    ### Parameters
    ----
    table_definition: str
        The TMDL-formatted string containing table definitions.
    ### Returns
    ----
    DataFrame:
        A pandas DataFrame with columns: table, name, type, data_type, expression.
    """
    current_table = None
    records = []
    lines = table_definition.splitlines()
    i = 0

    while i < len(lines):
        line = lines[i].strip()

        # Detect table
        table_match = re.match(r'^table\s+(.+)', line)
        if table_match:
            current_table = table_match.group(1).strip()
            i += 1
            continue

        # Detect measure
        measure_match = re.match(r"^measure\s+'?([^']+)'?\s*=\s*(.*)", line)
        if measure_match:
            measure_name = measure_match.group(1).strip()
            dax_expr = measure_match.group(2).strip()

            
            # If expression starts with triple backticks
            if dax_expr.startswith("```"):
                dax_lines = []
                i += 1
                while i < len(lines):
                    next_line = lines[i].strip()
                    if next_line == "```":
                        break
                    dax_lines.append(next_line)
                    i += 1
                dax_expr = "\n".join(dax_lines)
                i += 1  # Skip closing ```
            else:
                # Collect until next metadata or definition
                while i + 1 < len(lines):
                    peek = lines[i + 1].strip()
                    if peek.startswith(("formatString", "lineageTag", "annotation", "measure", "column", "table")):
                        break
                    i += 1
                    dax_expr += " " + lines[i].strip()


            records.append({
                'table': current_table,
                'name': measure_name,
                'type': 'measure',
                'data_type': None,
                'expression': dax_expr
            })
            i += 1
            continue

        # Detect column
        column_match = re.match(r"^column\s+'?([^'=]+?)'?(?:\s*=\s*(.+))?$", line)
        if column_match:
            col_name = column_match.group(1).strip()
            expression = column_match.group(2).strip() if column_match.group(2) else None
            data_type = None

            i += 1
            # Look ahead for dataType and other metadata
            while i < len(lines):
                next_line = lines[i].strip()
                if next_line.startswith("dataType:"):
                    data_type = next_line.replace("dataType:", "").strip()
                elif next_line.startswith(("lineageTag:", "formatString", "summarizeBy", "annotation", "dataCategory", "sourceColumn:")):
                    pass  # Ignore other metadata for now
                elif next_line.startswith(("column", "measure", "partition", "table")):
                    break
                i += 1

            records.append({
                'table': current_table,
                'name': col_name,
                'type': 'column',
                'data_type': data_type,
                'expression': expression
            })
            continue

        i += 1

    return pd.DataFrame(records)

def parse_measure(measure: Dict[str, Any]) -> str:
    """    Parses a measure definition into an HTML representation.
    ### Parameters  
    ----
    measure: dict
        A dictionary containing measure details with keys like 'name', 'expression', 'formatString', and 'description'.
    ### Returns
    ----
        str: An HTML representation of the measure.
    """
    expr = measure.get("expression", "")
    if isinstance(expr, list):
        expr = "\n".join(expr)
    return f"""
    <details class="measure-section">
      <summary>{measure.get('name')}</summary>
      <div><b>Expression:</b><pre>{expr}</pre></div>
      <div><b>Format:</b> {measure.get('formatString', '')}</div>
      <div><b>Description:</b> {measure.get('description', '')}</div>
    </details>
    """

def parse_partition(partition: Dict[str, Any]) -> str:
    """    Parses a partition definition into an HTML representation.
    ### Parameters
    ----
    partition: dict
        A dictionary containing partition details with keys like 'name', 'mode', 'queryGroup', and 'source'.
    ### Returns
    ----
        str: An HTML representation of the partition.
    """
    source = partition.get('source', {}).get('expression', '')
    if isinstance(source, list):
        source = "\n".join(source)
    return f"""
    <details class="partition-section">
      <summary>{partition.get('name')}</summary>
      <div><b>Mode:</b> {partition.get('mode', '')}</div>
      <div><b>Query Group:</b> {partition.get('queryGroup', '')}</div>
      <div><b>Source Expression:</b><pre class="language-js"><code class="language-js">{source}</code></pre></div>
    </details>
    """

def parse_role(role: Dict[str, Any]) -> str:
    """    Parses a role definition into an HTML representation.
    ### Parameters
    ----
    role: dict
        A dictionary containing role details with keys like 'name', 'modelPermission', and 'tablePermissions'.
    ### Returns
    ----
        str: An HTML representation of the role.
    """
    perms = role.get('modelPermission', '')
    table_perms = ""
    for tp in role.get("tablePermissions", []):
        if 'filterExpression' not in tp.keys():
            tp['filterExpression'] = "<em>No filter</em>"
        table_perms += f"<li>{tp['name']}: {tp['filterExpression']}</li>"
    return f"""
    <details class="role-section">
      <summary>{role.get('name')} <span class="badge">{perms}</span></summary>
      <ul>{table_perms}</ul>
    </details>
    """

def parse_columns_table(columns) -> str:
    """    Parses a list of column definitions into an HTML table.
    ### Parameters  
    ----
    columns: list
        A list of dictionaries, each representing a column with keys like 'name', 'dataType', 'formatString', and 'description'.
    ### Returns
    ----
        str: An HTML representation of the columns in a table format.   
    """
    if not columns:
        return "<em>No columns</em>"
    out = ['<table class="columns-table"><thead><tr><th>Name</th><th>Data Type</th><th>Format</th><th>Description</th></tr></thead><tbody>']
    for col in columns:
        out.append(
            f"<tr>"
            f"<td><b>{col.get('name')}</b></td>"
            f"<td>{col.get('dataType','')}</td>"
            f"<td>{col.get('formatString','')}</td>"
            f"<td>{col.get('description','')}</td>"
            f"</tr>"
        )
    out.append('</tbody></table>')
    return "\n".join(out)

def parse_table(table: Dict[str, Any]) -> str:
    """    Parses a table definition into an HTML representation.
    ### Parameters
    ----
    table: dict
        A dictionary containing table details with keys like 'name', 'columns', 'measures', and 'partitions'.
    ### Returns 
    ----
        str: An HTML representation of the table, including its columns, measures, and partitions.
    """
    columns_html = parse_columns_table(table.get("columns", []))
    measures_html = "".join([parse_measure(m) for m in table.get("measures", [])])
    partitions_html = "".join([parse_partition(p) for p in table.get("partitions", [])])
    anchor = f"table-{table.get('name').replace(' ', '_')}"
    return f"""
    <section class="table-card" id="{anchor}">
      <h2>{table.get('name')}</h2>
      <div class="table-section">
        <h3>Columns</h3>
        {columns_html}
        <details>
          <summary>Measures ({len(table.get('measures', []))})</summary>
          <div>{measures_html or "<em>No measures</em>"}</div>
        </details>
        <details>
          <summary>Partitions ({len(table.get('partitions', []))})</summary>
          <div>{partitions_html or "<em>No partitions</em>"}</div>
        </details>
      </div>
    </section>
    """

def parse_relationship(rel: Dict[str, Any]) -> str:
    """    Parses a relationship definition into an HTML representation.
    ### Parameters
    ----
    rel: dict
        A dictionary containing relationship details with keys like 'fromTable', 'fromColumn', 'toTable', 'toColumn', and 'cardinality'.
    ### Returns
    ----
        str: An HTML representation of the relationship, including the tables and columns involved, and the cardinality.
    """
    cardinality = rel.get('cardinality') or 'Unknown'
    return (f"<li><b>{rel['fromTable']}.{rel['fromColumn']}</b> "
            f"<span class='arrow'>&#8594;</span> "
            f"<b>{rel['toTable']}.{rel['toColumn']}</b> "
            f"<span class='cardinality'>[{cardinality}]</span></li>")

def parse_config(model: Dict[str, Any], bim: Dict[str, Any]) -> str:
    """    Parses the model configuration into an HTML representation.
    ### Parameters
    ----
    model: dict
        A dictionary containing model details with keys like 'culture', 'sourceQueryCulture', and 'annotations'.
    bim: dict
        A dictionary containing the entire BIM structure, which may include additional metadata.    
    ### Returns
    ----
        str: An HTML representation of the model configuration, including compatibility level, culture, and annotations.
    """
    compatibility_level = bim.get('compatibilityLevel') or model.get('compatibilityLevel', '')
    annos = model.get("annotations", [])
    annos_html = ""
    for a in annos:
        if a['name'] == "PBI_QueryOrder":
            continue  # skip long query order annotation
        annos_html += f"<li><b>{a['name']}</b>: {a['value']}</li>"
    return f"""
    <section class="config-section">
      <h2>Model Configuration</h2>
      <ul>
        <li><b>Compatibility Level:</b> {compatibility_level}</li>
        <li><b>Culture:</b> {model.get('culture', '')}</li>
        <li><b>Source Query Culture:</b> {model.get('sourceQueryCulture', '')}</li>
        {annos_html}
      </ul>
    </section>
    """

def generate_index_of_tables(tables) -> str:
    """    Generates an HTML representation of an index of tables, with links to each table's section.
    ### Parameters
    ----
    tables: list
        A list of dictionaries, each representing a table with a 'name' key.
    ### Returns
    ----
        str: An HTML representation of the index of tables, with links to each table's section.
    """
    items = []
    for t in tables:
        anchor = f"table-{t.get('name').replace(' ', '_')}"
        items.append(f'<li><a href="#{anchor}">{t.get("name")}</a></li>')
    return f"""
    <nav class="table-index">
      <h2>Tables Index</h2>
      <ul>
        {''.join(items)}
      </ul>
    </nav>
    """

def generate_cytoscape_diagram(tables, relationships):
    """    Generates an HTML representation of a data model diagram using Cytoscape.js.
    ### Parameters
    ----
    tables: list
        A list of dictionaries, each representing a table with a 'name' key.
    relationships: list
        A list of dictionaries, each representing a relationship with keys like 'fromTable', 'fromColumn', 'toTable', 'toColumn', and 'cardinality'.
    ### Returns
    ----
        str: An HTML representation of the data model diagram, including nodes for tables and edges for relationships.
    """ 
    cytoscape_nodes = []
    cytoscape_edges = []
    for t in tables:
        cytoscape_nodes.append({
            "data": {
                "id": t['name'],
                "label": t['name']
            }
        })
    for rel in relationships:
        cardinality = rel.get('cardinality') or "Unknown"
        cytoscape_edges.append({
            "data": {
                "id": f"{rel['fromTable']}__{rel['toTable']}__{rel['fromColumn']}",
                "source": rel['fromTable'],
                "target": rel['toTable'],
                "label": f"{rel['fromColumn']} → {rel['toColumn']} [{cardinality}]"
            }
        })
    js_nodes = json.dumps(cytoscape_nodes)
    js_edges = json.dumps(cytoscape_edges)
    return f"""
    <section class="diagram-section">
      <h2>Data Model Diagram</h2>
      <div id="model-diagram" style="width:100%;min-height:350px;height:470px;border:1px solid #e0e5ea;border-radius:8px;background:#fafcff;"></div>
      <script src="https://unpkg.com/cytoscape/dist/cytoscape.min.js"></script>
      <script>
        var cy = cytoscape({{
          container: document.getElementById('model-diagram'),
          elements: {{
            nodes: {js_nodes},
            edges: {js_edges}
          }},
          style: [
            {{
              selector: 'node',
              style: {{
                'label': 'data(label)',
                'background-color': '#0078d4',
                'color': '#fff',
                'font-size': '14px',
                'text-valign': 'center',
                'text-halign': 'center',
                'width': 95,
                'height': 95,
                'shape': 'round-rectangle',
                'border-width': 2,
                'border-color': '#004578',
                'text-wrap': 'ellipsis',
                'text-max-width': 80
              }}
            }},
            {{
              selector: 'edge',
              style: {{
                'curve-style': 'bezier',
                'width': 3,
                'target-arrow-shape': 'triangle',
                'target-arrow-color': '#005a9e',
                'line-color': '#79b8f3',
                'label': 'data(label)',
                'font-size': '12px',
                'color': '#444',
                'text-background-color': '#fff',
                'text-background-opacity': 1,
                'text-background-padding': '4px'
              }}
            }}
          ],
          layout: {{
            name: 'cose',
            animate: false,
            padding: 30
          }}
        }});
      </script>
      <p style="font-size:0.95em;color:#666;">(Drag nodes to rearrange. Relationships are labeled with key columns and cardinality.)</p>
    </section>
    """

def generate_bim_documentation(bim_json_text: str, output_html_path: str):
    """    Generates an HTML documentation for a Business Intelligence Model (BIM) based on its JSON representation.
    ### Parameters
    ----
    bim_json_text: str
        A string containing the JSON representation of the BIM, typically loaded from a .bim file.
    output_html_path: str
        The path where the HTML documentation will be saved.
    ### Returns
    ----
        None: The function writes the generated HTML to the specified output path.
    """
    bim = json.loads(bim_json_text)
    model = bim['model']
    tables = model.get("tables", [])
    relationships = model.get("relationships", [])
    roles = model.get("roles", [])

    config_html = parse_config(model, bim)
    diagram_html = generate_cytoscape_diagram(tables, relationships)
    relationships_html = "".join([parse_relationship(r) for r in relationships])
    # COLLAPSED BY DEFAULT
    relationships_section = f"""
    <details class="relationships-section">
      <summary><h2 style='display:inline'>Relationships</h2></summary>
      <ul>{relationships_html or "<li>No relationships</li>"}</ul>
    </details>
    """
    roles_html = "".join([parse_role(r) for r in roles])
    roles_section = f"""
    <section class="roles-section">
      <h2>Roles / RLS</h2>
      {roles_html or "<em>No roles defined</em>"}
    </section>
    """
    tables_index_html = generate_index_of_tables(tables)
    tables_html = "".join([parse_table(t) for t in tables])

    html = f"""
    <!DOCTYPE html>
    <html lang="en">
    <head>
      <meta charset="UTF-8">
      <link href="https://cdnjs.cloudflare.com/ajax/libs/prism/1.29.0/themes/prism.min.css" rel="stylesheet" />
      <script src="https://cdnjs.cloudflare.com/ajax/libs/prism/1.29.0/prism.min.js"></script>
      <title>Power BI Model Documentation</title>
      <meta name="viewport" content="width=device-width, initial-scale=1">
      <style>
        html,body {{
          font-family: 'Segoe UI', 'Roboto', Arial, sans-serif;
          background: #f3f6fa;
          margin: 0;
          color: #222;
        }}
        body {{
          max-width: 1100px;
          margin: 0 auto;
          padding: 0 0 4em 0;
        }}
        header {{
          background: linear-gradient(90deg,#0078d4 0,#00bcf2 100%);
          color: #fff;
          padding: 2rem 1.5rem 1.2rem 1.5rem;
          margin-bottom: 2rem;
          box-shadow: 0 2px 8px #0001;
        }}
        header h1 {{
          margin: 0;
          font-size: 2.5rem;
          letter-spacing: 1px;
        }}
        .diagram-section {{
          background: #fff;
          border-radius: 8px;
          box-shadow: 0 2px 8px #0001;
          margin-bottom: 2em;
          padding: 1em 1.5em;
        }}
        .table-index {{
          background: #fff;
          border-radius: 8px;
          box-shadow: 0 1px 4px #0001;
          margin-bottom: 2em;
          padding: 1em 1.5em;
        }}
        .table-index h2 {{
          font-size: 1.3em;
          margin-top: 0;
        }}
        .table-index ul {{
          columns: 2 200px;
          padding-left: 1.2em;
          margin: 0;
        }}
        .table-index li {{
          margin-bottom: 0.3em;
        }}
        .table-index a {{
          color: #0078d4;
          text-decoration: none;
        }}
        .table-index a:hover {{
          text-decoration: underline;
        }}
        section {{
          margin-bottom: 2em;
        }}
        .config-section, .relationships-section, .roles-section {{
          background: #fff;
          border-radius: 8px;
          box-shadow: 0 2px 8px #0001;
          margin-bottom: 2em;
          padding: 1em 1.5em;
        }}
        .table-card {{
          background: #fff;
          border-radius: 8px;
          box-shadow: 0 2px 8px #0001;
          margin-bottom: 2em;
          padding: 1.5em 1.5em 1em 1.5em;
        }}
        .table-card > h2 {{
          margin: 0 0 1em 0;
          color: #0078d4;
          font-size: 1.5em;
        }}
        .table-section h3 {{
          margin-top: 0.5em;
        }}
        .columns-table {{
          width: 100%;
          border-collapse: collapse;
          margin-bottom: 1em;
          background: #f7fbfe;
          border-radius: 5px;
          overflow-x: auto;
          font-size: 0.97em;
        }}
        .columns-table th, .columns-table td {{
          border: 1px solid #e0e5ea;
          padding: 0.5em 0.7em;
          text-align: left;
          vertical-align: top;
          word-break: break-word;
        }}
        .columns-table th {{
          background: #e5f3ff;
          color: #005a9e;
        }}
        .columns-table td {{
          background: #fafdff;
        }}
        .columns-table tr:hover td {{
          background: #f0f8ff;
        }}
        .table-section details {{
          margin-bottom: 0.8em;
          background: #f7fbfe;
          border-radius: 6px;
          box-shadow: 0 1px 3px #0078d411;
        }}
        .table-section summary {{
          font-weight: 500;
          font-size: 1.1em;
          padding: 0.5em 1em 0.5em 0.5em;
          cursor: pointer;
        }}
        .measure-section > summary,
        .partition-section > summary,
        .role-section > summary {{
          font-weight: 400;
          font-size: 1em;
          padding: 0.3em 0.7em;
        }}
        .measure-section[open] > summary,
        .partition-section[open] > summary,
        .role-section[open] > summary {{
          color: #005a9e;
        }}
        pre {{
          white-space: pre-wrap;
          word-break: break-all;
          background: #f3f6fa;
          padding: 0.5em;
          border-radius: 5px;
          font-size: 0.98em;
          overflow-x: auto;
        }}
        .arrow {{
          color: #888;
          font-size: 1.2em;
          margin: 0 0.5em;
        }}
        .cardinality {{
          background: #e5f3ff;
          color: #0078d4;
          border-radius: 4px;
          font-size: 0.92em;
          padding: 0.07em 0.5em;
          margin-left: 0.2em;
        }}
        .badge {{
          display: inline-block;
          background: #e5f3ff;
          color: #0078d4;
          border-radius: 5px;
          font-size: 0.85em;
          padding: 0.15em 0.6em;
          margin-left: 0.5em;
          vertical-align: middle;
        }}
        ul {{
          padding-left: 1.5em;
        }}
        details.relationships-section > summary {{
          cursor: pointer;
        }}
        details.relationships-section[open] > summary {{
          color: #0078d4;
        }}
        @media (max-width: 600px) {{
          body, header {{
            padding-left: 0.5em;
            padding-right: 0.5em;
          }}
          .table-card, .config-section, .relationships-section, .roles-section, .diagram-section, .table-index {{
            padding-left: 0.5em;
            padding-right: 0.5em;
          }}
          .table-index ul {{
            columns: 1 150px;
          }}
        }}
      </style>
    </head>
    <body>
      <header>
        <h1>Power BI Semantic Model Documentation</h1>
      </header>
      {config_html}
      {diagram_html}
      {relationships_section}
      {roles_section}
      {tables_index_html}
      {tables_html}
    </body>
    </html>
    """

    with open(output_html_path, "w", encoding="utf-8") as f:
        f.write(html)

def load_bim_file_as_string(filepath: str) -> str:
    """    Loads a BIM file and returns its contents as a string.
    ### Parameters
    ----    
    filepath: str
        The path to the BIM file to be loaded.
    ### Returns
    ----
        str: The contents of the BIM file as a string.
    """
    with open(filepath, "r", encoding="utf-8") as f:
        contents = f.read()

    # Optional: Validate it's proper JSON
    try:
        json.loads(contents)  # ensure it's valid
    except json.JSONDecodeError as e:
        raise ValueError(f"Invalid JSON in {filepath}: {e}")

    return contents


'''.
           @@@@@@@@@@
       @@@@..........@@@@
    @@@         .        @@@
  @@.           .         . @@
 @  .     _     .         .   @
@........| |...................@    *********************************************
@      . | |   _____  .        @
@      . | |  |  __ \ .        @    La Data Web
@      . | |__| |  | |.   ***  @
@........|____| |  | |...*   *.@    Copyright © 2024 Ignacio Barrau
@   .       . | |__| |. *     *@
@   .       . |_____/ . *     *@    *********************************************
@   .       .         . *     *@
@   .       .         . *******@
@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
'''

import json
import requests
from simplepbi import utils
import pandas as pd

class PushDatasets():
    """Simple library to use the Power BI api and obtain datasets from it.
    """

    def __init__(self, token):
        """Create a simplePBI object to request admin API
        Args:
            token: String
                Bearer Token to use the Power Bi Rest API
        """
        self.token = token
    
    def get_dataset_tables(self, dataset_id):
        """Returns a list of tables within the specified dataset from My workspace.
        ### Parameters
        ----
        dataset_id: str uuid
            The Power Bi Dataset id. You can take it from PBI Service URL
        ### Returns
        ----
        Dict:
            A dictionary containing tables from a dataset in My workspace.
        """
        try:
            url = "https://api.powerbi.com/v1.0/myorg/datasets/{}/tables".format(dataset_id)
            res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def get_dataset_tables_in_group(self, workspace_id, dataset_id):
        """Returns a list of tables within the specified dataset from the specified workspace.
        ### Parameters
        ----
        workspace_id: str uuid
            The Power Bi workspace id. You can take it from PBI Service URL
        dataset_id: str uuid
            The Power Bi Dataset id. You can take it from PBI Service URL
        ### Returns
        ----
        Dict:
            A dictionary containing tables from a dataset in the workspace.
        """
        try:
            url = "https://api.powerbi.com/v1.0/myorg/groups/{}/datasets/{}/tables".format(workspace_id, dataset_id)
            res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
      
    def delete_dataset_tablerows(self, dataset_id, table_name):
        """Deletes all rows from the specified table within the specified dataset from My workspace.
        ### Parameters
        ----
        dataset_id: str uuid
            The Power Bi Dataset id. You can take it from PBI Service URL
        table_name: str
            The table name
        ### Returns
        ----
        Response object from requests library. 201 OK
        
        """
        try: 
            url= "https://api.powerbi.com/v1.0/myorg/datasets/{}/tables/{}/rows".format(dataset_id, table_name)   
            headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)}
            res = requests.delete(url, headers=headers)
            res.raise_for_status()
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def delete_dataset_tablerows_in_group(self, workspace_id, dataset_id, table_name):
        """Deletes all rows from the specified table within the specified dataset from the specified workspace.
        ### Parameters
        ----
        workspace_id: str uuid
            The Power Bi Workspace id. You can take it from PBI Service URL
        dataset_id: str uuid
            The Power Bi Dataset id. You can take it from PBI Service URL
        table_name: str
            The table name
        ### Returns
        ----
        Response object from requests library. 201 OK
        
        """
        try: 
            url= "https://api.powerbi.com/v1.0/myorg/groups/{}/datasets/{}/tables/{}/rows".format(workspace_id, dataset_id, table_name)   
            headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)}
            res = requests.delete(url, headers=headers)
            res.raise_for_status()
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
 
            
    def post_dataset_preview(self, DefaultRetentionPolicy, name, tables, relationships=None):
        """Creates a new Streaming Dataset on My workspace. It's only for push datasets
        ### Parameters
        ----    
        DefaultRetentionPolicy: str
            The default retention policy. It can be basicFIFO or None
        ### Request Body
        ----
        name: str
            The dataset name
        tables: list dict
            The dataset tables. Example
            [{
              "name": "Product",
              "columns": [
                {
                  "name": "ProductID",
                  "dataType": "Int64"
                },
                {
                  "name": "Name",
                  "dataType": "string"
                }
                ]
             }]
        defaultMode: str
            The dataset mode or type. It is forced as PushStreaming only for now.
        relationshops:
            The dataset relationships. Please read the docs before using this parameter
        ### Returns
        ----
        Response object from requests library. 201 OK
        
        """
        try: 
            url= "https://api.powerbi.com/v1.0/myorg/datasets?defaultRetentionPolicy={}".format(DefaultRetentionPolicy)
            body = {
                "name": name,
                "tables": tables,
                "defaultMode": "PushStreaming"
            }
            if relationships != None:
                body["relationships"]=relationships
            headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)}            
            res = requests.post(url, data = json.dumps(body), headers = headers)
            res.raise_for_status()
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def post_dataset_in_group_preview(self, DefaultRetentionPolicy, workspace_id, name, tables, relationships=None):
        """Creates a new Streaming Dataset in the specified workspace. It's only for Push datasets
        ### Parameters
        ----    
        workspace_id: str uuid
            The Power Bi workspace id. You can take it from PBI Service URL 
        DefaultRetentionPolicy: str
            The default retention policy. It can be basicFIFO or None
        ### Request Body
        ----
        name: str
            The dataset name
        tables: list dict
            The dataset tables. Example
            [{
              "name": "Product",
              "columns": [
                {
                  "name": "ProductID",
                  "dataType": "Int64"
                },
                {
                  "name": "Name",
                  "dataType": "string"
                }
                ]
             }]
        defaultMode: str
            The dataset mode or type. It is forced as PushStreaming only for now.
        relationshops:
            The dataset relationships. Please read the docs before using this parameter
        ### Returns
        ----
        Response object from requests library. 201 OK
        
        """
        try: 
            url= "https://api.powerbi.com/v1.0/myorg/groups/{}/datasets?defaultRetentionPolicy={}".format(workspace_id, DefaultRetentionPolicy)
            body = {
                "name": name,
                "tables": tables,
                "defaultMode": "PushStreaming"
            }
            if relationships != None:
                body["relationships"]=relationships
            headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)}            
            res = requests.post(url, data = json.dumps(body), headers = headers)
            res.raise_for_status()
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def post_rows_table_dataset(self, dataset_id, table_name, rows):
        """Adds new data rows to the specified table within the specified dataset from My workspace.
        This API call only supports push datasets.
        ### Parameters
        ----    
        dataset_id: str uuid
            The Power Bi Dataset id. You can take it from PBI Service URL
        table_name: str
            The table name
        ### Request Body
        ----
        rows: object[]
            An array of data rows pushed to a dataset table. Each element is a collection of properties represented using key-value format. Example:
            [{
              "ProductID": 1,
              "Name": "Adjustable Race",
              "Category": "Components",
              "IsCompete": true,
              "ManufacturedOn": "07/30/2014"
            },
            {
              "ProductID": 2,
              "Name": "LL Crankarm",
              "Category": "Components",
              "IsCompete": true,
              "ManufacturedOn": "07/30/2014"
            }
            }]        
        ### Returns
        ----
        Response object from requests library. 200 OK
        
        """
        try: 
            url= "https://api.powerbi.com/v1.0/myorg/datasets/{}/tables/{}/rows".format(dataset_id, table_name)
            body = {
                "rows": rows
            }
            headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)}            
            res = requests.post(url, data = json.dumps(body), headers = headers)
            res.raise_for_status()
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def post_rows_table_dataset_in_group(self, workspace_id, dataset_id, table_name, rows):
        """Adds new data rows to the specified table within the specified dataset from the specified workspace.
        This API call only supports push datasets.
        ### Parameters
        ----    
        workspace_id: str uuid
            The Power Bi Workspace id. You can take it from PBI Service URL
        dataset_id: str uuid
            The Power Bi Dataset id. You can take it from PBI Service URL
        table_name: str
            The table name
        ### Request Body
        ----
        rows: object[]
            An array of data rows pushed to a dataset table. Each element is a collection of properties represented using key-value format. Example:
            [{
              "ProductID": 1,
              "Name": "Adjustable Race",
              "Category": "Components",
              "IsCompete": true,
              "ManufacturedOn": "07/30/2014"
            },
            {
              "ProductID": 2,
              "Name": "LL Crankarm",
              "Category": "Components",
              "IsCompete": true,
              "ManufacturedOn": "07/30/2014"
            }
            }] 
        ### Returns
        ----
        Response object from requests library. 200 OK
        
        """
        try: 
            url= "https://api.powerbi.com/v1.0/myorg/groups/{}/datasets/{}/tables/{}/rows".format(workspace_id, dataset_id, table_name)
            body = {
                "rows": rows
            }
            headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)}            
            res = requests.post(url, data = json.dumps(body), headers = headers)
            res.raise_for_status()
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def put_table_preview(self, dataset_id, table_name, columns, description, isHidden, measures=None, rows=None, source=None):
        """Updates the metadata and schema for the specified table within the specified dataset from My workspace.
        This request is in PREVIEW. It doesn't have measures, rows and sources parameters available'
        This API call only supports push datasets.
        ### Parameters
        ----    
        dataset_id: str uuid
            The Power Bi Dataset id. You can take it from PBI Service URL
        table_name: str
            The table name
        ### Request Body
        ----
        description: str
            The table description
        isHidden: bool
            Optional. Whether this dataset table is hidden
        name: str
            The table name
        columns: list dict
            	The column schema for this table
            The dataset tables. Example            
              "columns": [
                {
                  "name": "ProductID",
                  "dataType": "Int64"
                },
                {
                  "name": "Name",
                  "dataType": "string"
                }
                ]            
        ### Returns
        ----
        Response object from requests library. 201 OK
        
        """
        try: 
            url= "https://api.powerbi.com/v1.0/myorg/datasets/{}/tables/{}".format(dataset_id, table_name)
            body = {
                "name": table_name,
                "description": description,
                "isHidden": isHidden,
                "columns": columns
            }                
            '''
            if measures != None:
                body["measures"]=measures
            if rows != None:
                body["rows"]=rows
            if source != None:
                body["source"]=source
            '''
            headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)}            
            res = requests.put(url, data = json.dumps(body), headers = headers)
            res.raise_for_status()
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def put_table_in_group_preview(self, workspace_id, dataset_id, table_name, columns, description, isHidden, measures=None, rows=None, source=None):
        """Updates the metadata and schema for the specified table within the specified dataset from the specified workspace.
        This request is in PREVIEW. It doesn't have measures, rows and sources parameters available'
        This API call only supports push datasets.
        ### Parameters
        ----    
        workspace_id: str uuid
            The Power Bi Workspace id. You can take it from PBI Service URL
        dataset_id: str uuid
            The Power Bi Dataset id. You can take it from PBI Service URL
        table_name: str
            The table name
        ### Request Body
        ----
        tables: list dict
            The dataset tables. Example
            [{
              "name": "Product",
              "columns": [
                {
                  "name": "ProductID",
                  "dataType": "Int64"
                },
                {
                  "name": "Name",
                  "dataType": "string"
                }
                ]
             }]
        ### Returns
        ----
        Response object from requests library. 201 OK
        
        """
        try: 
            url= "https://api.powerbi.com/v1.0/myorg/groups/{}/datasets/{}/tables/{}".format(workspace_id, dataset_id, table_name)
            body = {
                "name": table_name ,
                "description": description,
                "isHidden": isHidden,
                "columns": columns
            }                
            '''
            if measures != None:
                body["measures"]=measures
            if rows != None:
                body["rows"]=rows
            if source != None:
                body["source"]=source
            '''
            headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)}            
            res = requests.put(url, data = json.dumps(body), headers = headers)
            res.raise_for_status()
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)


r'''.
           @@@@@@@@@@
       @@@@..........@@@@
    @@@         .        @@@
  @@.           .         . @@
 @  .     _     .         .   @
@........| |...................@    *********************************************
@      . | |   _____  .        @
@      . | |  |  __ \ .        @    La Data Web
@      . | |__| |  | |.   ***  @
@........|____| |  | |...*   *.@    Copyright © 2022 Ignacio Barrau
@   .       . | |__| |. *     *@
@   .       . |_____/ . *     *@    *********************************************
@   .       .         . *     *@
@   .       .         . *******@
@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
'''

import json
import requests
from simplepbi import utils
from datetime import date, timedelta
import io
import pandas as pd

class Admin():
    """Simple library to use the Power BI api and obtain datasets from it.
    """

    def __init__(self, token):
        """Create a simplePBI object to request admin API
        Args:
            token: String
                Bearer Token to use the Power Bi Rest API
        """
        self.token = token
    
    def get_datasets(self, filter=None, skip=None, top=None):
        """Returns a list of datasets for the organization..
        ### Parameters
        ----
        filter: string
            Filters the results based on a boolean condition
        skip: int 
            Skips the first n results. Use with top to fetch results beyond the first 5000.
        top: int
            Returns only the first n results. This parameter is mandatory and must be in the range of 1-5000.
        ### Returns
        ----
        Dict:
            A dictionary containing all the datasets in the tenant.
        """
        try:
            url = "https://api.powerbi.com/v1.0/myorg/admin/datasets?"
            if filter != None:
                url = url + "$filter={}".format(filter)
            if top != None:
                url = url + "&$top={}".format(top)
            if skip != None:
                url = url + "&$skip={}".format(skip) 
            res = requests.get(url.replace("?&", "?"), headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def get_datasets_in_group(self, workspace_id, expand=None, filter=None, skip=None, top=None):
        """Returns a list of datasets from the specified workspace.
        ### Parameters
        ----
        workspace_id:
            The Power Bi workspace id. You can take it from PBI Service URL
        expand: string
            Expands related entities inline, receives a comma-separated list of data types. Supported: users, reports, dashboards, datasets, dataflows, workbooks
        filter: string
            Filters the results based on a boolean condition
        skip: int 
            Skips the first n results. Use with top to fetch results beyond the first 5000.
        top: int
            Returns only the first n results. This parameter is mandatory and must be in the range of 1-5000.
        ### Returns
        ----
        Dict:
            A dictionary containing all the datasets in the workspace.
        """
        try:
            url = "https://api.powerbi.com/v1.0/myorg/admin/groups/{}/datasets?".format(workspace_id)
            if expand != None:
                url = url + "$expand={}".format(expand)
            if filter != None:
                url = url + "&$filter={}".format(filter)
            if skip != None:
                url = url + "&$skip={}".format(skip)   
            if top != None:
                url = url + "&$top={}".format(top)
            res = requests.get(url.replace("?&", "?"), headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def get_dataset_users(self, dataset_id):
        """Returns a list of users that have access to the specified dataset (Preview).
        ### Parameters
        ----
        dataset_id:
            The Power Bi Dataset id. You can take it from PBI Service URL
        ### Returns
        ----
        Dict:
            A dictionary containing all the users in the dataset.
        """
        try:
            url = "https://api.powerbi.com/v1.0/myorg/admin/datasets/{}/users".format(dataset_id)
            res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def get_datasources(self, dataset_id):
        """Returns a list of datasources for the specified dataset.
        ### Parameters
        ----
        dataset_id:
            The Power Bi Dataset id. You can take it from PBI Service URL
        ### Returns
        ----
        Dict:
            A dictionary containing all the datasources in the dataset.
        """
        try:
            url = "https://api.powerbi.com/v1.0/myorg/admin/datasets/{}/datasources".format(dataset_id)
            res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def get_dataset_to_dataflows_links_in_group(self, workspace_id):
        """Returns a list of upstream dataflows for datasets from the specified workspace.
        ### Parameters
        ----
        workspace_id:
            The Power Bi workspace id. You can take it from PBI Service URL
        ### Returns
        ----
        Dict:
            A dictionary containing all the upstream dataflows in the dataset from a workspace
        """
        try:
            url = "https://api.powerbi.com/v1.0/myorg/admin/groups/{}/datasets/upstreamDataflows".format(workspace_id)
            res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
        
    def get_reports(self, filter=None, skip=None, top=None):
        """Returns a list of reports for the organization.
        ### Parameters
        ----
        filter: string
            Filters the results based on a boolean condition
        skip: int 
            Skips the first n results. Use with top to fetch results beyond the first 5000.
        top: int
            Returns only the first n results. This parameter is mandatory and must be in the range of 1-5000.
        ### Returns
        ----
        Dict:
            A dictionary containing all the reports in the tenant.
        """
        try:
            url = "https://api.powerbi.com/v1.0/myorg/admin/reports?"
            if filter != None:
                url = url + "$filter={}".format(filter)
            if top != None:
                url = url + "&$top={}".format(top)
            if skip != None:
                url = url + "&$skip={}".format(skip)  
            res = requests.get(url.replace("?&", "?"), headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def get_reports_in_group(self, workspace_id, filter=None, skip=None, top=None):
        """Returns a list of reports from the specified workspace.
        ### Parameters
        ----
        workspace_id:
            The Power Bi workspace id. You can take it from PBI Service URL
        filter: string
            Filters the results based on a boolean condition
        skip: int 
            Skips the first n results. Use with top to fetch results beyond the first 5000.
        top: int
            Returns only the first n results. This parameter is mandatory and must be in the range of 1-5000.
        ### Returns
        ----
        Dict:
            A dictionary containing all the reports in the workspace.
        """
        try:
            url = "https://api.powerbi.com/v1.0/myorg/admin/groups/{}/reports?".format(workspace_id)
            if filter != None:
                url = url + "$filter={}".format(filter)
            if top != None:
                url = url + "&$top={}".format(top)
            if skip != None:
                url = url + "&$skip={}".format(skip) 
            res = requests.get(url.replace("?&", "?"), headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def get_reports_users(self, report_id):
        """Returns a list of users that have access to the specified report (Preview).
        ### Parameters
        ----
        report_id:
            The Power Bi Report id. You can take it from PBI Service URL
        ### Returns
        ----
        Dict:
            A dictionary containing all the users in the report.
        """
        try:
            url = "https://api.powerbi.com/v1.0/myorg/admin/reports/{}/users".format(report_id)
            res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def get_groups(self, top, expand=None, filter=None, skip=None):
        """Returns a workspace for the organization.
        ### Parameters
        ----
        top: int
            Returns only the first n results. This parameter is mandatory and must be in the range of 1-5000.
        expand: string
            Expands related entities inline, receives a comma-separated list of data types. Supported: users, reports, dashboards, datasets, dataflows, workbooks
        filter: string
            Filters the results based on a boolean condition
        skip: int 
            Skips the first n results. Use with top to fetch results beyond the first 5000.
        ### Returns
        ----
        Dict:
            A dictionary containing all the workspaces in the organization.
        """
        try:
            url = "https://api.powerbi.com/v1.0/myorg/admin/groups?$top={}".format(top)
            if expand != None:
                url = url + "&$expand={}".format(expand)
            if filter != None:
                url = url + "&$filter={}".format(filter)
            if skip != None:
                url = url + "&$skip={}".format(skip)                
            res = requests.get(url.replace("?&", "?"), headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def get_group(self, group_id, expand=None):
        """Returns a workspace for the organization.
        ### Parameters
        ----
        group_id: str
            The Power Bi Workspace id. You can take it from PBI Service URL
        expand: string
            Expands related entities inline, receives a comma-separated list of data types. Supported: users, reports, dashboards, datasets, dataflows, workbooks
        ### Returns
        ----
        Dict:
            A dictionary containing all the workspaces in the organization.
        """
        try:
            url = "https://api.powerbi.com/v1.0/myorg/admin/groups/{}".format(group_id)
            if expand != None:
                url = url + "?$expand={}".format(expand)              
            res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def get_groups_users(self, group_id):
        """Returns a list of users that have access to the specified workspace. This is a preview API call.
        ### Parameters
        ----
        group_id: str
            The Power Bi Workspace id. You can take it from PBI Service URL.
        ### Returns
        ----
        Dict:
            A dictionary containing all the users in the workspace.
        """
        try:
            url = "https://api.powerbi.com/v1.0/myorg/admin/groups/{}/users".format(group_id)
            res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def restore_deleted_group(self, workspace_id, emailAddress, name=None):
        """Restores a deleted workspace.
        This API call only supports restoring workspaces in the new workspace experience.
        ### Parameters
        ----
        workspace_id:
            The Power Bi workspace id. You can take it from PBI Service URL
        ### Request Body
        ----
        groupUserAccessRight: GroupUserAccessRight
            Access rights user has for the workspace (Permission level: Admin, Contributor, Member, Viewer or None). This is mandatory
        emailAddress: str
            Email address of the user. This is mandatory.
        name: str
            The name of the group to be restored.
        ### Returns
        ----
        Response object from requests library. 200 OK
        
        """
        try: 
            url= "https://api.powerbi.com/v1.0/myorg/admin/groups/{}/restore".format(workspace_id)
            body = {
                "emailAddress": emailAddress 
            }
            if name != None:
                body["name"]=name            
                
            headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)}
            
            res = requests.post(url, data = json.dumps(body), headers = headers)
            res.raise_for_status()
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)

    def get_dashboards(self, expand=None, filter=None, skip=None, top=None):
        """Returns a list of dashboards for the organization.
        ### Parameters
        ----
        expand: string
            Expands related entities inline, receives a comma-separated list of data types. Supported: users, reports, dashboards, datasets, dataflows, workbooks
        filter: string
            Filters the results based on a boolean condition
        skip: int 
            Skips the first n results. Use with top to fetch results beyond the first 5000.
        top: int
            Returns only the first n results. 
        ### Returns
        ----
        Dict:
            A dictionary containing all the dashboards in the tenant.
        """
        try:
            url = "https://api.powerbi.com/v1.0/myorg/admin/dashboards?"
            if expand != None:
                url = url + "$expand={}".format(expand)
            if filter != None:
                url = url + "&$filter={}".format(filter)
            if top != None:
                url = url + "&$top={}".format(top)
            if skip != None:
                url = url + "&$skip={}".format(skip)  
            res = requests.get(url.replace("?&", "?"), headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def get_dashboards_in_group(self, workspace_id, filter=None, skip=None, top=None):
        """Returns a list of dashboards from the specified workspace.
        ### Parameters
        ----
        workspace_id:
            The Power Bi workspace id. You can take it from PBI Service URL
        filter: string
            Filters the results based on a boolean condition
        skip: int 
            Skips the first n results. Use with top to fetch results beyond the first 5000.
        top: int
            Returns only the first n results. This parameter is mandatory and must be in the range of 1-5000.
        ### Returns
        ----
        Dict:
            A dictionary containing all the dashboards in the workspace.
        """
        try:
            url = "https://api.powerbi.com/v1.0/myorg/admin/groups/{}/dashboards?".format(workspace_id)
            if filter != None:
                url = url + "$filter={}".format(filter)
            if top != None:
                url = url + "&$top={}".format(top)
            if skip != None:
                url = url + "&$skip={}".format(skip) 
            res = requests.get(url.replace("?&", "?"), headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def get_dashboards_users(self, dashboard_id):
        """Returns a list of users that have access to the specified dashboard (Preview).
        ### Parameters
        ----
        dashboard_id:
            The Power Bi dashboard id. You can take it from PBI Service URL
        ### Returns
        ----
        Dict:
            A dictionary containing all the users in the dashboard.
        """
        try:
            url = "https://api.powerbi.com/v1.0/myorg/admin/dashboards/{}/users".format(dashboard_id)
            res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
           
    def get_tiles(self, dashboard_id):
        """Returns a list of tiles within the specified dashboard.
        ### Parameters
        ----
        dashboard_id:
            The Power Bi dashboard id. You can take it from PBI Service URL
        ### Returns
        ----
        Dict:
            A dictionary containing all the tiles in the dashboard.
        """
        try:
            url = "https://api.powerbi.com/v1.0/myorg/admin/dashboards/{}/tiles".format(dashboard_id)
            res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
    
    def get_dataflows(self, filter=None, skip=None, top=None):
        """Returns a list of dataflows for the organization.
        ### Parameters
        ----
        filter: string
            Filters the results based on a boolean condition
        skip: int 
            Skips the first n results. Use with top to fetch results beyond the first 5000.
        top: int
            Returns only the first n results. This parameter is mandatory and must be in the range of 1-5000.
        ### Returns
        ----
        Dict:
            A dictionary containing all the dataflows in the tenant.
        """
        try:
            url = "https://api.powerbi.com/v1.0/myorg/admin/dataflows?"
            if filter != None:
                url = url + "$filter={}".format(filter)
            if top != None:
                url = url + "&$top={}".format(top)
            if skip != None:
                url = url + "&$skip={}".format(skip)  
            res = requests.get(url.replace("?&", "?"), headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def get_dataflows_in_group(self, workspace_id, filter=None, skip=None, top=None):
        """Returns a list of dataflows from the specified workspace.
        ### Parameters
        ----
        workspace_id:
            The Power Bi workspace id. You can take it from PBI Service URL
        filter: string
            Filters the results based on a boolean condition
        skip: int 
            Skips the first n results. Use with top to fetch results beyond the first 5000.
        top: int
            Returns only the first n results. This parameter is mandatory and must be in the range of 1-5000.
        ### Returns
        ----
        Dict:
            A dictionary containing all the dataflows in the workspace.
        """
        try:
            url = "https://api.powerbi.com/v1.0/myorg/admin/groups/{}/dataflows?".format(workspace_id)
            if filter != None:
                url = url + "$filter={}".format(filter)
            if top != None:
                url = url + "&$top={}".format(top)
            if skip != None:
                url = url + "&$skip={}".format(skip) 
            res = requests.get(url.replace("?&", "?"), headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def get_dataflows_users(self, dataflow_id):
        """Returns a list of users that have access to the specified dataflow (Preview).
        ### Parameters
        ----
        dataflow_id:
            The Power Bi dataflow id. You can take it from PBI Service URL
        ### Returns
        ----
        Dict:
            A dictionary containing all the users in the dataflow.
        """
        try:
            url = "https://api.powerbi.com/v1.0/myorg/admin/dataflows/{}/users".format(dataflow_id)
            res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
        
    def get_dataflow_datasources(self, dataflow_id):
        """Returns a list of datasources for the specified dataflow.
        ### Parameters
        ----
        dataflow_id:
            The Power Bi Dataflow id. You can take it from PBI Service URL
        ### Returns
        ----
        Dict:
            A dictionary containing all the datasources in the dataflow.
        """
        try:
            url = "https://api.powerbi.com/v1.0/myorg/admin/dataflows/{}/datasources".format(dataflow_id)
            res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def get_upstream_dataflows_in_group(self, workspace_id, dataflow_id):
        """Returns a list of upstream dataflows for the specified dataflow.
        ### Parameters
        ----
        workspace_id:
            The Power Bi workspace id. You can take it from PBI Service URL
        dataflow_id:
            The Power Bi Dataflow id. You can take it from PBI Service URL
        ### Returns
        ----
        Dict:
            A dictionary containing all the dataflows in the workspace.
        """
        try:
            url = "https://api.powerbi.com/v1.0/myorg/admin/groups/{}/dataflows/{}/upstreamDataflows".format(workspace_id, dataflow_id)
            res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)    
            
    def export_dataflow(self, dataflow_id):
        """Exports the definition for the specified dataflow to a JSON file.
        ### Parameters
        ----
        dataflow_id:
            The Power Bi Dataflow id. You can take it from PBI Service URL
        ### Returns
        ----
        Dict:
            A dictionary containing all the metadata built in the dataflow.
        """
        try:
            url = "https://api.powerbi.com/v1.0/myorg/admin/dataflows/{}/export".format(dataflow_id)
            res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def get_apps(self, top):
        """Returns a list of apps for the organization.
        ### Parameters
        ----
        top: int
            Returns only the first n results. This parameter is mandatory and must be in the range of 1-5000.
        ### Returns
        ----
        Dict:
            A dictionary containing all the apps in the tenant.
        """
        try:
            url = "https://api.powerbi.com/v1.0/myorg/admin/apps?$top={}".format(top)
            res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def get_apps_users(self, app_id):
        """Returns a list of users that have access to the specified app (Preview).
        ### Parameters
        ----
        app_id:
            The Power Bi app id. You can take it from PBI Service URL
        ### Returns
        ----
        Dict:
            A dictionary containing all the users in the app.
        """
        try:
            url = "https://api.powerbi.com/v1.0/myorg/admin/apps/{}/users".format(app_id)
            res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def get_capacities(self, expand=None):
        """Returns a list of capacities for the organization.
        ### Parameters
        ----
        expand: int
            Expands related entities inline
        ### Returns
        ----
        Dict:
            A dictionary containing all the capacities in the tenant.
        """
        try:
            url = "https://api.powerbi.com/v1.0/myorg/admin/capacities?"
            if expand != None:
                url = url + "$expand={}".format(expand)
            res = requests.get(url, headers={'Content-Type': 'capacitielication/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def get_capacities_users(self, capacity_id):
        """Returns a list of users that have access to the specified capacitie (Preview).
        ### Parameters
        ----
        capacity_id:
            The Power Bi capacity id. You can take it from PBI Service URL
        ### Returns
        ----
        Dict:
            A dictionary containing all the users in the capacity.
        """
        try:
            url = "https://api.powerbi.com/v1.0/myorg/admin/capacities/{}/users".format(capacity_id)
            res = requests.get(url, headers={'Content-Type': 'capacitielication/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def get_refreshable_capacity(self, capacity_id, refreshable_id, expand=None):
        """Returns the specified refreshable for the specified capacity that the user has access to.
        Power BI retains a seven-day refresh history for each dataset, up to a maximum of sixty refreshes.
        ### Parameters
        ----
        capacity_id:
            The Power Bi capacity id. You can take it from PBI Service URL
        refreshable_id
            The Power Bi refreshable id in the capacity.
        expand: int
            Expands related entities inline, receives a comma-separated list of data types. Supported: capacities and groups
        ### Returns
        ----
        Dict:
            A dictionary containing all the refreshables in the capacity.
        """
        try:
            url = "https://api.powerbi.com/v1.0/myorg/admin/capacities/{}/refreshables/{}?".format(capacity_id, refreshable_id)
            if expand != None:
                url = url + "$expand={}".format(expand)
            res = requests.get(url, headers={'Content-Type': 'capacitielication/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def get_refreshables_capacity(self, capacity_id, top, expand=None, filter=None, skip=None):
        """Returns a list of refreshables for the specified capacity that the user has access to.
        Power BI retains a seven-day refresh history for each dataset, up to a maximum of sixty refreshes.
        ### Parameters
        ----
        capacity_id:
            The Power Bi capacitie id. You can take it from PBI Service URL
        top: int
            Returns only the first n results. This parameter is mandatory and must be in the range of 1-5000.
        expand: string
            Expands related entities inline, receives a comma-separated list of data types. Supported: users, reports, dashboards, datasets, dataflows, workbooks
        filter: string
            Filters the results based on a boolean condition
        skip: int 
            Skips the first n results. Use with top to fetch results beyond the first 5000.
        ### Returns
        ----
        Dict:
            A dictionary containing all the users in the capacity.
        """
        try:
            url = "https://api.powerbi.com/v1.0/myorg/admin/capacities/{}/refreshables?$top={}".format(capacity_id, top) 
            if expand != None:
                url = url + "&$expand={}".format(expand)
            if filter != None:
                url = url + "&$filter={}".format(filter)
            if skip != None:
                url = url + "&$skip={}".format(skip) 
            res = requests.get(url.replace("?&", "?"), headers={'Content-Type': 'capacitielication/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
           
    def get_user_artifact_access_preview(self, userGraphId, return_pandas=False):
        '''Returns a list of artifacts that the given user have access to (Preview).
        *** THIS REQUEST IS IN PREVIEW IN SIMPLEPBI ***
        
        ### Parameters
        ----
        userGraphId: str uuid
            The graph ID of user
        return_pandas: bool
            Flag to specify if you want to return a dict response or a pandas dataframe of events. By default pandas
        ### Returns
        ----
        If return_pandas = True returns a Pandas dataframe concatenating iterations otherwise it returns a dict of the response
        ### Limitations
        ----
        Maximum 200 requests per hour.
        '''        
        columnas = ['artifactId', 'displayName', 'artifactType', 'accessRight']
        df_total = pd.DataFrame(columns=columnas)
        dict_total = {}
        list_total = []
        url = "https://api.powerbi.com/v1.0/myorg/admin/users/{}/artifactAccess".format(userGraphId)
        headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)}
        ban = True   
        contar = 0    
        try:
            while(ban):        
                res = requests.get(url, headers=headers)
                if return_pandas:
                    js = json.dumps(res.json()["ArtifactAccessEntities"])
                    df = pd.read_json(js)
                    df_total = pd.concat([df_total, df], sort=True, ignore_index=True)
                    print("Building dataframe iteration: ", str(contar))
                else:
                    if res.json()["ArtifactAccessEntities"]:
                        list_total.extend(res.json()["ArtifactAccessEntities"])
                        #append_value(dict_total, "ArtifactAccessEntities", res.json()["ArtifactAccessEntities"])
                        print("Building dict iteration: ", str(contar))
                print(res.status_code)
                contar = contar +1    
                if "continuationUri" not in res.json():
                    ban=False
                else:
                    url = res.json()["continuationUri"]                   
                    print(res.json()["continuationUri"])    
            if return_pandas:
                print(df_total.tail())
                return df_total
            else:
                dict_total = {'ArtifactAccessEntities': list_total }
                return dict_total
        except requests.exceptions.Timeout:
            print("ERROR: The request method has exceeded the Timeout")
        except requests.exceptions.TooManyRedirects:
            print("ERROR: Bad URL try a different one")
        except requests.exceptions.RequestException as e:
            print("Catastrophic error.")
            raise SystemExit(e)
        except Exception as ex:
            print("ERROR: ", ex)
            
    def get_unused_artifacts(self, workspace_id):
        """Returns a list of artifacts from the specified workspace with last used date.
        ### Parameters
        ----
        workspace_id:
            The Power Bi workspace id. You can take it from PBI Service URL
        ### Returns
        ----
        Dict:
            A dictionary containing all the artifacts in the workspace.
        """
        try:
            url = "https://api.powerbi.com/v1.0/myorg/admin/groups/{}/unused".format(workspace_id)
            res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)        
            
    def get_pipelines(self, expand=None, filter=None, skip=None, top=None):
        """Returns a list of pipelines for the organization.
        ### Parameters
        ----
        expand: string
            Expands related entities inline, receives a comma-separated list of data types. Supported: users, reports, pipelines, datasets, dataflows, workbooks
        filter: string
            Filters the results based on a boolean condition
        skip: int 
            Skips the first n results. Use with top to fetch results beyond the first 5000.
        top: int
            Returns only the first n results. This parameter is mandatory and must be in the range of 1-5000.
        ### Returns
        ----
        Dict:
            A dictionary containing all the pipelines in the tenant.
        """
        try:
            url = "https://api.powerbi.com/v1.0/myorg/admin/pipelines?"
            if expand != None:
                url = url + "$expand={}".format(expand)
            if filter != None:
                url = url + "&$filter={}".format(filter)
            if top != None:
                url = url + "&$top={}".format(top)
            if skip != None:
                url = url + "&$skip={}".format(skip)  
            res = requests.get(url.replace("?&", "?"), headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
			
    def get_pipelines_users(self, pipeline_id):
        """Returns a list of users that have access to the specified pipeline (Preview).
        ### Parameters
        ----
        pipeline_id:
            The Power Bi pipeline id. You can take it from PBI Service URL
        ### Returns
        ----
        Dict:
            A dictionary containing all the users in the pipeline.
        """
        try:
            url = "https://api.powerbi.com/v1.0/myorg/admin/pipelines/{}/users".format(pipeline_id)
            res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)    
        
    def get_imports(self, expand=None, filter=None, skip=None, top=None):
        """Returns a list of imports for the organization.
        ### Parameters
        ----
        expand: string
            Expands related entities inline, receives a comma-separated list of data types. Supported: users, reports, imports, datasets, dataflows, workbooks
        filter: string
            Filters the results based on a boolean condition
        skip: int 
            Skips the first n results. Use with top to fetch results beyond the first 5000.
        top: int
            Returns only the first n results. This parameter is mandatory and must be in the range of 1-5000.
        ### Returns
        ----
        Dict:
            A dictionary containing all the imports in the tenant.
        """
        try:
            url = "https://api.powerbi.com/v1.0/myorg/admin/imports?"
            if expand != None:
                url = url + "$expand={}".format(expand)
            if filter != None:
                url = url + "&$filter={}".format(filter)
            if top != None:
                url = url + "&$top={}".format(top)
            if skip != None:
                url = url + "&$skip={}".format(skip)  
            res = requests.get(url.replace("?&", "?"), headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
    
    def get_refreshables(self, expand=None, filter=None, skip=None, top=None):
        """Returns a list of refreshables for the organization.
        ### Parameters
        ----
        top: int
            Returns only the first n results. This parameter is mandatory and must be in the range of 1-5000.
        expand: string
            Expands related entities inline, receives a comma-separated list of data types. Supported: users, reports, refreshables, datasets, dataflows, workbooks
        filter: string
            Filters the results based on a boolean condition
        skip: int 
            Skips the first n results. Use with top to fetch results beyond the first 5000.
        ### Returns
        ----
        Dict:
            A dictionary containing all the refreshables in the tenant.
        """
        try:
            url = "https://api.powerbi.com/v1.0/myorg/admin/capacities/refreshables?$top={}".format(top)
            if expand != None:
                url = url + "&$expand={}".format(expand)
            if filter != None:
                url = url + "&$filter={}".format(filter)            
            if skip != None:
                url = url + "&$skip={}".format(skip)  
            res = requests.get(url.replace("?&", "?"), headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
        
    def get_encryption_keys(self, expand=None, filter=None, skip=None, top=None):
        """Returns the encryption keys for the tenant.
        ### Parameters
        ----
        ### Returns
        ----
        Dict:
            A dictionary containing all the refreshables in the tenant.
        """
        try:
            url = "https://api.powerbi.com/v1.0/myorg/admin/tenantKeys"
            res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def add_encryption_key_preview(self, activate, isDefault, keyVaultKeyIdentifier, name):
        """Adds an encryption key for Power BI workspaces assigned to a capacity.
        *** THIS REQUEST IS IN PREVIEW IN SIMPLEPBI ***
        ### Parameters
        ----
        ### Request Body
        ----
        All the keys are requested for the body
        activate: bool
            Indicates to activate any inactivated capacities to use this key for its encryption.
        isDefault: bool
            Whether an encryption key is the default key for the entire tenant. Any newly created capacity inherits the default key.
        keyVaultKeyIdentifier: str
            The URI that uniquely specifies an encryption key in Azure Key Vault.
        name:
            The name of the encryption key        
        ### Returns
        ----
        Response object from requests library. 200 OK
        
        """
        try: 
            url= "https://api.powerbi.com/v1.0/myorg/admin/tenantKeys"
            body = {
                "activate": activate, 
                "isDefault": isDefault,
                "keyVaultKeyIdentifier":keyVaultKeyIdentifier,
                "name":name
            }
                
            headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)}
            
            res = requests.post(url, data = json.dumps(body), headers = headers)
            res.raise_for_status()
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def rotate_encryption_key_preview(self, tenantKeyId, keyVaultKeyIdentifier):
        """Adds an encryption key for Power BI workspaces assigned to a capacity.
        *** THIS REQUEST IS IN PREVIEW IN SIMPLEPBI ***
        ### Parameters
        ----
        tenantKeyId: str uuid
            The tenant key ID
        ### Request Body
        ----        
        keyVaultKeyIdentifier: str
            The URI that uniquely specifies an encryption key in Azure Key Vault.
        ### Returns
        ----
        Response object from requests library. 200 OK
        
        """
        try: 
            url= "https://api.powerbi.com/v1.0/myorg/admin/tenantKeys/{}/Default.Rotate".format(tenantKeyId)
            body = {
                "keyVaultKeyIdentifier":keyVaultKeyIdentifier
            }
                
            headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)}
            
            res = requests.post(url, data = json.dumps(body), headers = headers)
            res.raise_for_status()
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
    
    def add_user_to_group(self, workspace_id, groupUserAccessRight, emailAddress, displayName=None, graphId=None, identifier=None, principalType=None):
        """Grants user permissions to the specified workspace.
        This API call only supports updating workspaces in the new workspace experience and adding a user principle.
        ### Parameters
        ----
        workspace_id:
            The Power Bi workspace id. You can take it from PBI Service URL
        ### Request Body
        ----
        groupUserAccessRight: GroupUserAccessRight
            Access rights user has for the workspace (Permission level: Admin, Contributor, Member, Viewer or None). This is mandatory
        emailAddress: str
            Email address of the user. This is mandatory.
        displayName: str
            Display name of the principal
        graphId: str
            Identifier of the principal in Microsoft Graph. Only available for admin APIs
        identifier: str
            Object ID of the principal
        principalType: principalType
            The principal type (App, Group, User or None)
        ### Returns
        ----
        Response object from requests library. 200 OK
        
        """
        try: 
            url= "https://api.powerbi.com/v1.0/myorg/admin/groups/{}/users".format(workspace_id)
            body = {
                "groupUserAccessRight": groupUserAccessRight, 
                "emailAddress": emailAddress 
            }
            if displayName != None:
                body["diplayName"]=displayName
            if graphId != None:
                body["graphId"] = graphId
            if identifier != None:
                body["identifier"] = identifier
            if principalType != None:
                body["principalType"] = principalType
                
            headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)}
            
            res = requests.post(url, data = json.dumps(body), headers = headers)
            res.raise_for_status()
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def delete_user_from_group(self, workspace_id, user):
        """Removes user permissions from the specified workspace.
        This API call only supports updating workspaces in the new workspace experience and adding a user principle.
        ### Parameters
        ----
        workspace_id: str uuid
            The Power Bi workspace id. You can take it from PBI Service URL
        user: str
            The user principal name (UPN) of the user to remove (usually the user's email).
        ### Returns
        ----
        Response object from requests library. 200 OK
        
        """
        try: 
            url= "https://api.powerbi.com/v1.0/myorg/admin/groups/{}/users/{}".format(workspace_id, user)   
            headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)}
            res = requests.delete(url, headers=headers)
            res.raise_for_status()
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def update_group_preview(self, workspace_id, capacityId=None, dashboards=None, dataflowStorageId=None, dataflows=None, datasets=None, description=None, isOnDedicatedCapacity=None, isReadOnly=None, name=None, pipelineId=None, reports=None, state=None, typee=None, users=None, workbooks=None):    
        """Updates the properties of the specified workspace.
        *** THIS REQUEST IS IN PREVIEW IN SIMPLEPBI ***
        This API call call only updates workspaces in the new workspace experience. Only the name and description can be updated. The name must be unique inside an organization.
        ### Parameters
        ----
        workspace_id:
            The Power Bi workspace id. You can take it from PBI Service URL
        ### Request Body
        ----
        id: string
            The workspace ID
        capacityId: string
            The capacity ID
        dashboards: str[]
            List of the dashboards ids that belong to the group. Available only for admin API calls.
        dataflowStorageId: string
            The Power BI dataflow storage account ID
        dataflows: str[]
            List of the dataflows ids that belong to the group. Available only for admin API calls.
        datasets: str[]
            List of the datasets ids that belong to the group. Available only for admin API calls.
        description: string
            The group description. Available only for admin API calls.
        isOnDedicatedCapacity: bool
            Is the group on dedicated capacity
        isReadOnly: bool
            Is the group read only
        name: string
            The group name
        pipelineId: string
            The deployment pipeline ID that the workspace is assigned to. Available only for workspaces in the new workspace experience and only for admin API calls.
        reports: str[]
            List of the reports ids that belong to the group. Available only for admin API calls.
        state: string
            The group state. Available only for admin API calls.
        typee: string
            The type of group. Available only for admin API calls.
        users: GroupUser[]
            List of the users that belong to the group, and their access rights. This value will be empty. It will be removed from the payload response in an upcoming release. To retrieve user information on an artifact, please consider using the Get Group User APIs, or the PostWorkspaceInfo API with the getArtifactUser parameter.
        workbooks: str[]
            List of the workbooks ids that belong to the group. Available only for admin API calls.
        ### Returns
        ----
        Response object from requests library. 200 OK
        
        """
        try: 
            url= "https://api.powerbi.com/v1.0/myorg/admin/groups/{}".format(workspace_id)
            body = {
                "id": workspace_id
            }
            
            if capacityId != None:
                body["capacityId"]=capacityId
            if dashboards != None:
                body["dashboards"] = dashboards
            if dataflowStorageId != None:
                body["dataflowStorageId"] = dataflowStorageId
            if dataflows != None:
                body["dataflows"] = dataflows
            if datasets != None:
                body["datasets"]=datasets
            if description != None:
                body["description"] = description
            if isOnDedicatedCapacity != None:
                body["isOnDedicatedCapacity"] = isOnDedicatedCapacity
            if isReadOnly != None:
                body["isReadOnly"] = isReadOnly
            if name != None:
                body["name"]=name
            if pipelineId != None:
                body["pipelineId"] = pipelineId
            if reports != None:
                body["reports"] = reports
            if state != None:
                body["state"] = state
            if typee != None:
                body["type"] = typee
            if users != None:
                body["users"] = users
            if workbooks != None:
                body["workbooks"] = workbooks
                
            headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)}
            
            res = requests.patch(url, json.dumps(body), headers = headers)
            res.raise_for_status()
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def update_user_in_pipeline(self, pipeline_id, identifier, principalType, accessRight):
        """Grants user permissions to a specified deployment pipeline.
        
        ### Parameters
        ----
        pipeline_id:
            The Power Bi Deployment Pipeline id. You can take it from PBI Service URL
        ### Request Body
        ----
        identifier: str
            For Principal type 'User' provide UPN , otherwise provide Object ID of the principal. This is mandatory
        principalType: principalType
            The principal type (App, Group, User or None). This is mandatory.
        accessRight: GroupUserAccessRight
            accessRequired - Access rights a user has for the deployment pipeline. (Permission level: Admin). This is mandatory
        ### Returns
        ----
        Response object from requests library. 200 OK
        
        """
        try: 
            url= "https://api.powerbi.com/v1.0/myorg/admin/pipelines/{}/users".format(pipeline_id)
            body = {
                "identifier": identifier ,
                "principalType": principalType ,
                "accessRight": accessRight
            }
                
            headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)}
            
            res = requests.post(url, data = json.dumps(body), headers = headers)
            res.raise_for_status()
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def delete_user_from_pipeline(self, pipeline_id, identifier):
        """Removes user permissions from a specified deployment pipeline.
        
        ### Parameters
        ----
        pipeline_id: str uuid
            The deployment pipeline ID
        identifier: str
            For Principal type 'User' provide UPN , otherwise provide Object ID of the principal
        ### Returns
        ----
        Response object from requests library. 200 OK
        
        """
        try: 
            url= "https://api.powerbi.com/v1.0/myorg/admin/pipelines/{}/users/{}".format(pipeline_id, identifier)   
            headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)}
            res = requests.delete(url, headers=headers)
            res.raise_for_status()
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def assign_workspaces_to_capacity_preview(self, tagetCapacityObjectId, workspacesToAssign):
        """Assigns the specified workspaces to the specified Premium capacity.
        *** THIS REQUEST IS IN PREVIEW IN SIMPLEPBI ***
        ### Parameters
        ----
        ### Request Body
        ----
        capacityMigrationAssignments: Assignment contract for migrating workspaces to premium capacity as tenant admin
        
        targetCapacityObjectId: str
            The premium capacity ID
        workspacesToAssign: str[]
            List of the workspace IDs to be migrated to premium capacity
        ### Returns
        ----
        Response object from requests library. 200 OK
        
        """
        try: 
            url= "https://api.powerbi.com/v1.0/myorg/admin/capacities/AssignWorkspaces"
            body ={
                "tagetCapacityObjectId":tagetCapacityObjectId,
                "workspacesToAssign":workspacesToAssign
            }
            body_option2 ={
                "capacityMigrationAssignments":[{
                    "tagetCapacityObjectId":tagetCapacityObjectId,
                    "workspacesToAssign":workspacesToAssign
                }]
            }
                
            headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)}
            
            res = requests.post(url, data = json.dumps(body), headers = headers)
            res.raise_for_status()
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def unassign_workspaces_from_capacity_preview(self, workspacesToAssign):
        """Unassigns the specified workspaces from capacity.
        *** THIS REQUEST IS IN PREVIEW IN SIMPLEPBI ***
        ### Parameters
        ----
        ### Request Body
        ----        
        workspacesToAssign: str[]
            List of the workspace IDs to be migrated to premium capacity
        ### Returns
        ----
        Response object from requests library. 200 OK
        
        """
        try: 
            url= "https://api.powerbi.com/v1.0/myorg/admin/capacities/UnassignWorkspaces"
            body ={
                "workspacesToAssign":workspacesToAssign
            }
                
            headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)}
            
            res = requests.post(url, data = json.dumps(body), headers = headers)
            res.raise_for_status()
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
                                                                                         
    def get_activity_events_preview(self, activity_date=None, return_pandas=False, filter_event=None):
        '''Returns a dict of pandas dataframe of audit activity events for a tenant.
        *** THIS REQUEST IS IN PREVIEW IN SIMPLEPBI ***
        The continuation token is automtaically used to get all the results in the date.
        ### Parameters
        ----
        activity_date: str "yyyy-mm-dd"
            The Single date to get events from the whole day.
            If the date is not specify it will return yesterday events by default.
        return_pandas: bool
            Flag to specify if you want to return a dict response or a pandas dataframe of events.
        filter_event: query str
            Filters the results based on a boolean condition, using 'Activity', 'UserId', or both properties. Supports only 'eq' and 'and' operators.
            Ej: filter_event = "UserId eq 'ibarrau@ladataweb.com.ar' and Activity eq 'GetRefreshHistory'"
        ### Returns
        ----
        If return_pandas = True returns a Pandas dataframe concatenating iterations otherwise it returns a dict of the response
        ### Limitations
        ----
        Maximum 200 requests per hour.
        '''        
        columnas = ['Activity', 'ActivityId', 'AppId', 'AppName', 'AppReportId',
       'ArtifactId', 'ArtifactKind', 'ArtifactName', 'CapacityId',
       'CapacityName', 'ClientIP', 'ConsumptionMethod', 'CreationTime',
       'DataConnectivityMode', 'DatasetId', 'DatasetName',
       'DistributionMethod', 'FolderAccessRequests', 'FolderDisplayName',
       'FolderObjectId', 'GatewayClusters', 'HasFullReportAttachment', 'Id',
       'ImportDisplayName', 'ImportId', 'ImportSource', 'ImportType',
       'IsSuccess', 'IsTenantAdminApi', 'ItemName', 'LastRefreshTime',
       'ModelsSnapshots', 'ObjectId', 'Operation', 'OrganizationId',
       'RecordType', 'RefreshType', 'ReportId', 'ReportName', 'ReportType',
       'RequestId', 'TableName', 'UserAgent', 'UserId', 'UserKey', 'UserType',
       'WorkSpaceName', 'Workload', 'WorkspaceId']
        df_total = pd.DataFrame(columns=columnas)
        dict_total = {'activityEventEntities': [] }
        list_total = []
        if activity_date == None:
            activity_date = date.today()- timedelta(days=1)
        else:
            activity_date = date(int(activity_date.split("-")[0]),int(activity_date.split("-")[1]),int(activity_date.split("-")[2]))
        start = activity_date.strftime("'%Y-%m-%dT%H:%M:00.000Z'")
        end = activity_date.strftime("'%Y-%m-%dT23:59:59.000Z'")
        url = "https://api.powerbi.com/v1.0/myorg/admin/activityevents?startDateTime={}&endDateTime={}".format(start, end)
        if filter_event != None:
            url = url + "&$filter={}".format(filter_event)
        ban = True   
        contar = 0    
        try:
            print("Getting activity events for date: ", activity_date, "... running iterations...")
            while(ban):
                res = requests.get(url,
                    headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)}
                    )
                res.raise_for_status()
                if return_pandas:
                    js = json.dumps(res.json()["activityEventEntities"])
                    df = pd.read_json(io.StringIO(js))
                    if not df.empty and not df.isna().all(axis=None):
                        if not df_total.empty:
                            df_total = pd.concat([df_total, df], sort=True, ignore_index=True)
                        else:
                            df_total = df.copy()
                    #print("Building dataframe iteration: ", str(contar))
                else:
                    if res.json()["activityEventEntities"]:                
                        #append_value(dict_total, "activityEventEntities", res.json()["activityEventEntities"][0])
                        list_total.extend(res.json()["activityEventEntities"])
                        #print("Building dict iteration: ", str(contar))
                    
                #print(res.status_code)
                contar = contar +1
                #print(res.json()["continuationUri"])
                
                if res.json()["lastResultSet"] == True or res.json()["continuationUri"] == None:
                    ban=False
                url = res.json()["continuationUri"]
                #print("last result set: ", res.json()["lastResultSet"])
            print("Total iterations: ", contar)
            if return_pandas:
                return df_total
            else:
                dict_total = {'activityEventEntities': list_total }
                return dict_total
        except requests.exceptions.Timeout:
            print("ERROR: The request method has exceeded the Timeout")
        except requests.exceptions.TooManyRedirects:
            print("ERROR: Bad URL try a different one")
        except requests.exceptions.RequestException as e:
            print("Catastrophic error:", e)
            raise SystemExit(e)
        except Exception as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.text)
            
    def get_activity_events_last_28_days_preview(self, filter_event=None):
        '''Returns a pandas dataframe of audit activity events for the last 28 days at the tenant.
        *** THIS can take a several minutes because it loops 28 days and pagination ***
        *** THIS REQUEST IS IN PREVIEW IN SIMPLEPBI ***
        The continuation token is automtaically used to get all the results in the last 28 days starting yesterday.
        ### Parameter
        ----
        filter_event: query str
            Filters the results based on a boolean condition, using 'Activity', 'UserId', or both properties. Supports only 'eq' and 'and' operators.
            Ej: filter_event = "UserId eq 'ibarrau@ladataweb.com.ar' and Activity eq 'GetRefreshHistory'"        
        ### Returns
        ----
        Returns a Pandas dataframe concatenating iterations
        ### Limitations
        ----
        Maximum 200 requests per hour.
        '''

        # Getting last 30 days
        today = date.today() - timedelta(days=1)
        last_30_days = []
        for i in range(27):
            last_30_days.append((today - timedelta(days=i)).strftime("%Y-%m-%d"))

        # Creating an empty DataFrame with the structure needed    
        df = pd.DataFrame(['Activity', 'ActivityId', 'AppId', 'AppName', 'AppReportId',
           'ArtifactId', 'ArtifactKind', 'ArtifactName', 'CapacityId',
           'CapacityName', 'ClientIP', 'ConsumptionMethod', 'CreationTime',
           'DataConnectivityMode', 'DatasetId', 'DatasetName',
           'DistributionMethod', 'FolderAccessRequests', 'FolderDisplayName',
           'FolderObjectId', 'GatewayClusters', 'HasFullReportAttachment', 'Id',
           'ImportDisplayName', 'ImportId', 'ImportSource', 'ImportType',
           'IsSuccess', 'IsTenantAdminApi', 'ItemName', 'LastRefreshTime',
           'ModelsSnapshots', 'ObjectId', 'Operation', 'OrganizationId',
           'RecordType', 'RefreshType', 'ReportId', 'ReportName', 'ReportType',
           'RequestId', 'TableName', 'UserAgent', 'UserId', 'UserKey', 'UserType',
           'WorkSpaceName', 'Workload', 'WorkspaceId'])
        # Loop last 30 days appending the result in a single dataframe
        for i in last_30_days:
            if filter_event == None:    
                df_temp = self.get_activity_events_preview(i, return_pandas=True)
            else:
                df_temp = self.get_activity_events_preview(i, return_pandas=True, filter_event=filter_event)
            df = pd.concat([df, df_temp])
        if 0 in df.columns:
            # Drop 0 column of activities auto generated
            df = df.drop(columns=[0])
        # Remove NaN empty values of each type of activities auto generated
        df = df.dropna(subset=['Activity'])
        return df
            
    def get_modified_workspaces_preview(self, excludePersonalWorkspaces=True, modifiedSince=None):
        """Gets a list of workspace IDs in the organization. This is a preview API call.
        *** THIS REQUEST IS IN PREVIEW IN SIMPLEPBI ***
        ### Parameters
        ----
        excludePersonalWorkspaces: bool
            Whether to exclude personal workspaces
        modifiedSince: str-datetime
            format %Y-%m-%dT%H:%M:00.000Z
            
        ### Returns
        ----
        Dict:
            Returns a list of list that contains groups of maximum 100 workspaces.
        """
        lista_total = []
        contador = 100
        #datetime.strptime("2021-01-01 01:55:19", "%Y-%m-%d %H:%M:%S")
        #modify_date = modifiedSince.strftime("'%Y-%m-%dT%H:%M:00.000Z'")
        try:
            url = "https://api.powerbi.com/v1.0/myorg/admin/workspaces/modified?excludePersonalWorkspaces={}".format(excludePersonalWorkspaces)
            if modifiedSince != None:
                url = url + "&modifiedSince={}".format(modifiedSince)
            res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res = res.json()
            lista = [res[i]['id'] for i in range(len(res))]
            for item in range(len(lista)):
                if lista[item*100:item*100+100] != []:
                    lista_total.append(lista[item*100:item*100+100])
                else:
                    break
            return lista_total 
        
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
        
    def post_workspace_info(self, workspaces, lineage=True, datasourceDetails=True, datasetSchema=True, datasetExpressions=True, getArtifactUsers=True):
        """Initiates a call to receive metadata for the requested list of workspaces. This is a preview API call.
        *** THIS REQUEST IS IN PREVIEW IN SIMPLEPBI ***
        ### Parameters
        ----
        ### Request Body
        ----        
        workspaces: str[]
            List of the workspace IDs to ask for scan (it can't contain more than 100 workspaces)
        ### Returns
        ----
        Scan id in uuid format. 202 OK
        
        """
        try: 
            url= "https://api.powerbi.com/v1.0/myorg/admin/workspaces/getInfo?lineage={}&datasourceDetails={}&datasetSchema={}&datasetExpressions={}&getArtifactUsers={}".format(lineage, datasourceDetails, datasetSchema, datasetExpressions, getArtifactUsers)
            body ={
                "workspaces":workspaces
            }
                
            headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)}
            
            res = requests.post(url, data = json.dumps(body), headers = headers)
            res.raise_for_status()
            return res.json()["id"]
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)            
            
    def get_scan_status_preview(self, scan_id):
        """Gets the scan status for the specified scan. This is a preview API call.
            *** THIS REQUEST IS IN PREVIEW IN SIMPLEPBI ***
        ### Parameters
        ----
        scan_id: str uui
            The scan id obtained from posting workspaces info        
        ### Returns
        ----
        str:
            Returns the status of the scan. Succeeded means you are ready to request scans.
        """
        try:
            url = "https://api.powerbi.com/v1.0/myorg/admin/workspaces/scanStatus/{}".format(scan_id)
            res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()["status"]
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)            
            
    def get_scan_result_preview(self, scan_id):
        """Gets the scan result for the specified scan. This is a preview API call.
            *** THIS REQUEST IS IN PREVIEW IN SIMPLEPBI ***
        ### Parameters
        ----
        scan_id: str uui
            The scan id obtained from posting workspaces info        
        ### Returns
        ----
        str:
            Returns the status of the scan. Succeeded means you are ready to request scans.
        """
        try:
            url = "https://api.powerbi.com/v1.0/myorg/admin/workspaces/scanResult/{}".format(scan_id)
            res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def get_available_features(self):
        """
        Returns a list of available features for the user.
        ### Returns
        -------
        200 ok. A dict with the features.
        """
        try:
            url = "https://api.powerbi.com/v1.0/myorg/availableFeatures"
            res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)  
            
    def get_orphan_dataflows_preview(self):
        """Returns a list of all dataflows that are not used by a dataset.
            *** THIS REQUEST IS IN PREVIEW IN SIMPLEPBI ***
        ### Limitations
        ----
        It can only be used for organizations with less than 200 workspaces that contains dataflows. The PBI Rest API won't allow more than 200 requests in an hour.
        ### Returns
        ----
        List:
            A list containing all the dataflows without a dataset connected.
        """
        try:
            res_wp = self.get_groups(top=5000, expand="dataflows")
            workspaces = [res_wp["value"][i]["id"] for i in range(len(res_wp["value"])) if res_wp["value"][i]["dataflows"] != []]
            
            if len(workspaces) > 200:            
                return "You can't use this request because you have more than 200 workspaces (limitation)."
        
            #url_df = "https://api.powerbi.com/v1.0/myorg/admin/dataflows"
            #res_df = requests.get(url_df, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            #res_df.raise_for_status()
            res_df = self.get_dataflows()
            dataflows = [res_df["value"][i]["objectId"] for i in range(len(res_df["value"]))]
            
            actives = []
            orphans = []
            
            for wp in workspaces:
                url = "https://api.powerbi.com/v1.0/myorg/admin/groups/{}/datasets/upstreamDataflows".format(wp)
                res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
                if res.text != '' or res.status_code != 200:
                    actives.extend( [res.json()["value"][i]["dataflowObjectId"] for i in range(len(res.json()["value"])) ] )
                
            for df in dataflows:
                if df not in actives:
                    orphans.append(df)
            
            return orphans
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def get_dashboard_subscriptions_preview(self, dashboard_id):
        """Returns a list of subscriptions along with subscribees that the dashboard subscribed to. This is a preview API call.
        ### Parameters
        ----
        dashboard_id: str uui
            The dashboard id
        ### Returns
        ----
        Dict:
            Returns detailed subscriptions of the dashboard.
        ### Limitations
        ----
        Maximum 200 requests per hour.
        """
        try:
            url = "https://api.powerbi.com/v1.0/myorg/admin/dashboards/{}/subscriptions".format(dashboard_id)
            res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def get_report_subscriptions_preview(self, report_id):
        """Returns a list of subscriptions along with subscribees that the report subscribed to. This is a preview API call.
        ### Parameters
        ----
        report_id: str uui
            The report id
        ### Returns
        ----
        Dict:
            Returns detailed subscriptions of the report.
        ### Limitations
        ----
        Maximum 200 requests per hour.
        """
        try:
            url = "https://api.powerbi.com/v1.0/myorg/admin/reports/{}/subscriptions".format(report_id)
            res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def get_user_subscriptions_preview(self, user_id):
        """Returns a list of subscriptions that the given user has subscribed to (preview).
        ### Parameters
        ----
        user_id: str uui
            The graph ID or UPN of user
        ### Returns
        ----
        Dict:
            Returns  detailed subscriptions of the user.
        ### Limitations
        ----
        Maximum 200 requests per hour.
        """
        ban = True 
        url = "https://api.powerbi.com/v1.0/myorg/admin/users/{}/subscriptions".format(user_id)
        contar = 0
        list_total = []
        dict_total = {'SubscriptionEntities':[]}
        try:
            while(ban):                
                res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
                res.raise_for_status()
                
                if res.json()["SubscriptionEntities"]:
                    list_total.extend(res.json()["SubscriptionEntities"])
                    print("Building dict iteration: ", str(contar))
                    
                contar = contar +1
                print("Pagination: ", str(contar))
                
                if res.json()["continuationUri"] == None or res.json()["SubscriptionEntities"]==[]:
                    ban=False
                url = res.json()["continuationUri"]   
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def get_widely_shared_artifacts_links_shared_to_whole_organization(self):
        """Returns a list of artifacts shared to the whole organization through links.
        ### Returns
        ----
        Dict:
            Returns artifacts shared to the whole organization.
        ### Limitations
        ----
        Maximum 200 requests per hour.
        """
        ban = True 
        url = "https://api.powerbi.com/v1.0/myorg/admin/widelySharedArtifacts/linksSharedToWholeOrganization"
        contar = 0
        list_total = []
        dict_total = {'ArtifactAccessEntities':[]}
        try:
            while(ban):                
                res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
                res.raise_for_status()
                
                if res.json()["ArtifactAccessEntities"]:
                    list_total.extend(res.json()["ArtifactAccessEntities"])
                    print("Building dict iteration: ", str(contar))
                    
                contar = contar +1
                print("Pagination: ", str(contar))
                
                try:
                    res.json()["continuationUri"]                    
                except Exception as ex:
                    break
                    
                if res.json()["continuationUri"] == None or res.json()["ArtifactAccessEntities"]==[]:
                    ban=False        
                url = res.json()["continuationUri"]   
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def get_widely_shared_artifacts_published_to_web(self):
        """Returns a list of artifacts shared through published to web.
        ### Returns
        ----
        Dict:
            Returns artifacts published to web.
        ### Limitations
        ----
        Maximum 200 requests per hour.
        """
        ban = True 
        url = "https://api.powerbi.com/v1.0/myorg/admin/widelySharedArtifacts/publishedToWeb"
        contar = 0
        list_total = []
        dict_total = {'ArtifactAccessEntities':[]}
        try:
            while(ban):                
                res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
                res.raise_for_status()
                
                if res.json()["ArtifactAccessEntities"]:
                    list_total.extend(res.json()["ArtifactAccessEntities"])
                    print("Building dict iteration: ", str(contar))
                    
                contar = contar +1
                print("Pagination: ", str(contar))
                
                try:
                    res.json()["continuationUri"]                    
                except Exception as ex:
                    break
                    
                if res.json()["continuationUri"] == None or res.json()["ArtifactAccessEntities"]==[]:
                    ban=False        
                url = res.json()["continuationUri"]   
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def get_tenant_settings(self):
        """Returns a list of the tenant settings.
        ### Returns
        ----
        Dict:
            Response 200. A dict with tenant settings.
        ### Limitations
        ----
        Maximum 200 requests per hour.
        """
        try:
            url = "https://api.powerbi.com/v1/admin/tenantsettings"                     
            res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def get_datasets_roles_in_groups(self, workspace_id_list):
        """Returns a list of workspaces and datasets with their roles (RLS). It uses Scanner API
        ### Parameters
        ----
        workspace_id_list: uuid str[]
            The Power Bi Workspace id. You can take it from PBI Service URL like ['xxx-xxx-xxx-xxx', 'yyy-yyy-yyy-yyy']
        ### Limitations
        ----
        The list of workspaces can't contain more than 100 workspaces.        
        ### Returns
        ----
        Dict:
            A dictionary containing all the roles in datasets in workspaces.
        """
        lista_workspaces=[]
        lista_datasets=[]
        
        try:
            scan_id = self.post_workspace_info(workspace_id_list, datasetSchema=True)
            ssta = self.get_scan_status_preview(scan_id)
            res = self.get_scan_result_preview(scan_id)
                
            for w in range(len(res['workspaces'])):
                for d in range(len(res['workspaces'][w]['datasets'])):
                    if "roles" in res['workspaces'][w]['datasets'][d].keys():
                        item = [{ 
                            "id": res['workspaces'][w]['datasets'][d]['id'],
                            "name": res['workspaces'][w]['datasets'][d]['name'],
                            "roles": res['workspaces'][w]['datasets'][d]['roles'] 
                            }]
                        lista_datasets.extend(item)
                datasets = [{
                    "id": res['workspaces'][w]['id'],
                    "name": res['workspaces'][w]['name'],
                    "datasets": lista_datasets 
                    }]
                lista_workspaces.extend(datasets)
                lista_datasets=[]
            workspaces = {"workspaces": lista_workspaces}
            
            return workspaces
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def get_dataset_roles_in_group(self, workspace_id, dataset_id):
        """Return a list of roles (RLS) at a dataset from a workspace. It uses Scanner API
        ### Parameters
        ----
        workspace_id: uuid 
            The Power Bi Workspace id. You can take it from PBI Service URL
        dataset_id: uuid
            The Power Bi Semantic model id. You can take it from PBI Service URL
        ### Limitations
        ----
        The list of workspaces can't contain more than 100 workspaces.        
        ### Returns
        ----
        Dict:
            A dictionary containing all the roles in datasets in workspaces.
        """
        lista_workspaces=[]
        lista_datasets=[]
        workspace_id_list = [workspace_id]
        
        try:
            scan_id = self.post_workspace_info(workspace_id_list, datasetSchema=True)
            ssta = self.get_scan_status_preview(scan_id)
            res = self.get_scan_result_preview(scan_id)
            
            dataset = [ item for item in res['workspaces'][0]['datasets'] if item['id']==dataset_id ][0]
                
            if "roles" in dataset.keys():
                item = [{ 
                    "id": dataset['id'],
                    "name": dataset['name'],
                    "roles": dataset['roles'] 
                    }]
                lista_datasets.extend(item)
                datasets = [{
                    "id": res['workspaces'][0]['id'],
                    "name": res['workspaces'][0]['name'],
                    "datasets": lista_datasets 
                    }]
                lista_workspaces.extend(datasets)
                lista_datasets=[]
            else:
                lista_workspaces = "The dataset in the workspace doesn't have created roles."
            workspaces = {"workspaces": lista_workspaces}
            
            return workspaces
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)


'''.
           @@@@@@@@@@
       @@@@..........@@@@
    @@@         .        @@@
  @@.           .         . @@
 @  .     _     .         .   @
@........| |...................@    *********************************************
@      . | |   _____  .        @
@      . | |  |  __ \ .        @    La Data Web
@      . | |__| |  | |.   ***  @
@........|____| |  | |...*   *.@    Copyright © 2022 Ignacio Barrau
@   .       . | |__| |. *     *@
@   .       . |_____/ . *     *@    *********************************************
@   .       .         . *     *@
@   .       .         . *******@
@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
'''

import json
import requests

class Scorecards():
    """Simple library to use the Power BI api and obtain scorecards from it.
    """

    def __init__(self, token):
        """Create a simplePBI object to request scorecards API
        Args:
            token: String
                Bearer Token to use the Power Bi Rest API
        """
        self.token = token    
            
    def get_scorecard_in_group(self, workspace_id, scorecard_id, expand=None):
        """Returns the specified scorecard from the specified workspace.
        ### Parameters
        ----
        workspace_id: str uuid
            The Power Bi workspace id. You can take it from PBI Service URL
        scorecard_id: str uuid
            The Power Bi scorecard id.
        expand: str
            Accepts a comma-separated list of data types, which will be expanded inline in the response. Supports { goals, goalValues, aggregations, and notes }
        ### Returns
        ----
        Dict:
            A dictionary containing a scorecard in the workspace.
        """
        try:
            url = "https://api.powerbi.com/v1.0/myorg/groups/{}/scorecards({})".format(workspace_id, scorecard_id)
            if expand != None:
                url = url + "?$expand={}".format(expand)
            res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def get_scorecards_in_group(self, workspace_id, top=None):
        """Returns a list of scorecards from the specified workspace.
        ### Parameters
        ----
        workspace_id: str uuid
            The Power Bi workspace id. You can take it from PBI Service URL
        top: int
            Returns only the first n results. This parameter is mandatory and must be in the range of 1-5000.
        ### Returns
        ----
        Dict:
            A dictionary containing all the scorecards in the workspace.
        """
        try:
            url = "https://api.powerbi.com/v1.0/myorg/groups/{}/scorecards".format(workspace_id)            
            if top != None:
                url = url + "?$top={}".format(top)
            res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def get_scorecard_by_report(self, workspace_id, report_id, expand=None):
        """Reads a scorecard associated with an internal report ID.
        ### Parameters
        ----
        workspace_id: str uuid
            The Power Bi workspace id. You can take it from PBI Service URL
        report_id: str uuid
            The ID of the internal report associated with the scorecard
        expand: str
            Accepts a comma-separated list of data types, which will be expanded inline in the response. Supports { goals, goalValues, and aggregations }
        ### Returns
        ----
        Dict:
            A dictionary containing a scorecard in the workspace.
        """
        try:
            url = "https://api.powerbi.com/v1.0/myorg/groups/{}/scorecards/GetScorecardByReportId(reportId={})".format(workspace_id, report_id)
            if expand != None:
                url = url + "?$expand={}".format(expand)
            res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def post_scorecard(self, workspace_id, name, description=None, sensitivityLabelId=None):
        """Creates a new scorecard.
        ### Parameters
        ----
        workspace_id: str uuid
            The Power Bi workspace id. You can take it from PBI Service URL
        ### Request Body
        ----
        name: str
            The scorecard name
        description: str
            Optional. The scorecard description.
        sensitivityLabelId: str
            Optional. The GUID of a sensitivity label. If you don't want to select a sensitivity label, use a null or empty GUID (00000000-0000-0000-0000-000000000000). If default labels are enabled and/or enforced, they will be applied on the scorecard and dataset.
        ### Returns
        ----
        Dict:
            Response 200. A dict with a new report id.
        """
        try:
            url = "https://api.powerbi.com/v1.0/myorg/groups/{}/scorecards".format(workspace_id)
            body = {"name": name}
            if description != None:
                body["description"]=description
            if sensitivityLabelId != None:
                body["sensitivityLabelId"]=sensitivityLabelId
                
            headers = {'Content-Type': 'multipart/form-data', "Authorization": "Bearer {}".format(self.token)}
            res = requests.post(url, data = json.dumps(body), headers=headers)
            res.raise_for_status()
            
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def delete_scorecard_in_group(self, workspace_id, scorecard_id):
        """Deletes a scorecard by its ID.
        ### Parameters
        ----
        workspace_id: str uuid
            The Power Bi workspace id. You can take it from PBI Service URL
        scorecard_id: str uuid
            The Power Bi scorecard id.
        ### Returns
        ----
        Dict:
            A dictionary containing a scorecard in the workspace.
        """
        try:
            url = "https://api.powerbi.com/v1.0/myorg/groups/{}/scorecards({})".format(workspace_id, scorecard_id)
            res = requests.delete(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)

'''
    def move_goals_preview(self, workspace_id, scorecard_id, goalToMove, newNext=None, newParent=None, newPrevious=None):
        """Moves goals within the scorecard. Changes their ranks and parents.
        ### Parameters
        ----
        workspace_id: str uuid
            The Power Bi workspace id. You can take it from PBI Service URL
        scorecard_id: str uuid
            The unique identifier of the scorecard
        ### Request Body
        ----
        goalToMove: dict { currentParentId:"str uuid", goalId:"str uuid" }
            The rank validation information for the goal to be moved. The caller provides validation information to confirm that they know the existing position of a goal within the hierarchy of goals.
        newNext: str
            Optional. The rank validation information for the new next-sibling of the goal to be moved. The caller provides validation information to confirm that they know the existing position of a goal within the hierarchy of goals.
        newParent: str
            Optional. The rank validation information for the new previous-sibling of the goal to be moved. The caller provides validation information to confirm that they know the existing position of a goal within the hierarchy of goals.
        newPrevious: str
            
        ### Returns
        ----
        Dict:
            Response 204. 
        """
        try:
            url = "https://api.powerbi.com/v1.0/myorg/groups/{}/scorecards({})/MoveGoals()".format(workspace_id, scorecard_id)
            body = {"goalToMove": goalToMove}
            if newNext != None:
                body["newNext"]=newNext
            if newParent != None:
                body["newParent"]=newParent
            if newPrevious != None:
                body["newPrevious"]=newPrevious
                
            headers = {'Content-Type': 'multipart/form-data', "Authorization": "Bearer {}".format(self.token)}
            res = requests.post(url, data = json.dumps(body), headers=headers)
            res.raise_for_status()
            
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def patch_goals_preview(self, workspace_id, scorecard_id, goalToMove, newNext=None, newParent=None, newPrevious=None):
        """Updates a scorecard by its ID
        ### Parameters
        ----
        workspace_id: str uuid
            The Power Bi workspace id. You can take it from PBI Service URL
        scorecard_id: str uuid
            The unique identifier of the scorecard
        ### Request Body
        ----
        columnSettings: ScorecardColumnSetting[]	
            The display settings for columns on a scorecard
        createdTime: str
            The UTC time at creation
        datasetId: str
            The ID of the dataset associated with the scorecard
        description: str
            The scorecard description
        goals	Goal[]	The scorecard goals
        groupId: str
            The ID of the workspace
        id: str
            The scorecard ID
        lastModifiedTime: str
            The UTC time at last modification
        name: str
            The scorecard name
        permissions: ScorecardPermission	
            The scorecard permissions
        provisioningStatus: str
            Provisioning status of the scorecard. Supported { Initialized	, Completed, Failed, Deprovisioning, Deleted	 }
        reportId: str
            The ID of the internal report associated with the scorecard            
        ### Returns
        ----
        Dict:
            Response 200. 
        """
        try:
            url = "https://api.powerbi.com/v1.0/myorg/groups/{}/scorecards({})".format(workspace_id, scorecard_id)
            body = {"goalToMove": goalToMove}
            if newNext != None:
                body["newNext"]=newNext
            if newParent != None:
                body["newParent"]=newParent
            if newPrevious != None:
                body["newPrevious"]=newPrevious
            if newNext != None:
                body["newNext"]=newNext
            if newParent != None:
                body["newParent"]=newParent
            if newPrevious != None:
                body["newPrevious"]=newPrevious
                
            headers = {'Content-Type': 'multipart/form-data', "Authorization": "Bearer {}".format(self.token)}
            res = requests.patch(url, data = json.dumps(body), headers=headers)
            res.raise_for_status()
            
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
'''